/* ----------------------------------------------------------------
 *
 * andorgui.c  
 * 
 * Project: Andor Camera software (OCIW, Pasadena, CA)
 *
 * v0.001  2016-10-24  Christoph C. Birk, birk@carnegiescience.edu
 * v0.018  2016-11-14  Guider Class
 * v0.022  2016-11-21  new GUI layout
 * v0.030  2016-12-07  gwin experiments 
 * v0.040  2017-01-20  binX,Y -> binning
 * v0.042  2017-01-25  "send" command
 * v0.044  2017-01-26  var. number of message lines, Led
 * v0.051  2017-02-10  qltool: fix rev_line() "static"
 * v0.054  2017-09-07  SH-optics tests
 * v0.061  2018-01-23  telio (TCSIS protocol)
 * v0.062  2018-02-03  tests @ Clay (139.229.109.60)
 * v0.064  2018-02-09  tests @ Clay
 * v0.066  2018-02-23  add 'dt', apa=1, sendHost, lockfile
 * v0.067  2018-04-02  remove the lock-file on error at startup. 
 *                     mm-2 moves the guider probes and the telescope. 
 * v0.068  2018-04-xx  disable 'Q' in release version
 * v0.069  2018-04-26  add TCP/IP data server
 * FORK IT
 * v0.200  2018-05-09  andorserver
 * v0.203  2018-05-14  USB->TCP
 * v0.219  2018-08-06  MIKE run 
 * v0.220  2018-08-21  auto-start 'server' if 'localhost'
 * v0.221  2018-08-28  use '-i' switch for auto-start 'localhost'
 * v0.222  2020-10-17  horizontal layout
 * v0.223  2021-05-24  guider box in upper left corner at startup
 * v0.224  2022-07-06  "aeg" command, '-v' switch
 *
 * http://www.lco.cl/telescopes-information/magellan/
 *   operations-homepage/magellan-control-system/magellan-code/gcam
 *
 * http://confluence.lco.cl:8090/display/MATOP/
 *   Instrument-telescope+communication#_shack_hartmann_images
 *
 * IDEA add command history
 * IDEA scroll message log
 * IDEA EDS data stream 
 * IDEA "control" key required optionally
 * IDEA display 'sm'
 * IDEA dark current subtraction
 *
 * TODO .ini file (sky parity and PA)
 * TODO update FITS header (EPOCH wrong in Magellan example)
 *
 * ---------------------------------------------------------------- */

#ifndef DEBUG
#define DEBUG           1          /* debug level */
#endif

#define TIME_TEST       1

/* ------------------------------------------------------- */

#define _REENTRANT

#include <stdlib.h>                    /* atoi(),exit() */
#include <stdio.h>                     /* sprintf() */
#include <string.h>                    /* strcpy(),memset(),memcpy() */
#include <math.h>                      /* sqrt() */
#include <ctype.h>
#include <assert.h>
#include <errno.h>

#if (TIME_TEST > 0)
#include <limits.h> 
#include <sys/times.h>
#include <unistd.h>
#endif

#include <cxt.h>                       /* Xlib stuff */

#include "andor.h"
#include "tcpip.h"                     /* TCPIP stuff */
#include "utils.h"                     /* generic utilities */
#include "ptlib.h"                     /* POSIX threads lib */

#include "andorTCP.h"
#include "guider.h"                    /* qltool.h,graph.h,fits.h */
#include "gcpho.h"
#include "telio.h"
#include "random.h"

/* DEFINEs -------------------------------------------------------- */

#undef  TEST_GUIDER 
#define PA_BOX

#define P_TITLE         "Andor"

#define LOCKFILE        "/tmp/andorgui.lock"  // TODO TCPIP lock

#define MY_HELP         "andor.html"

#define DBE_DATAPATH    "datapath"
#define FMT_RUN         "run%d"
#define DBE_PREFIX      "prefix"    
#define DEF_PREFIX      "gdr"

#define DEF_HELP "http://instrumentation.obs.carnegiescience.edu/Software/Andor"
#if defined(MACOSX)
#define DEF_BROWSER     "open"         /* Safari */
#else
#define DEF_BROWSER     "firefox"
#endif

#define DEF_FONT        "lucidatypewriter"

#define TCPIP_PORT      (50001+(100*PROJECT_ID))  /* client port */

#define PREFUN          __func__

enum guider_modes_enum { MODE_NONE,MODE_PR,MODE_SH,MODE_SV };

/* EXTERNs -------------------------------------------------------- */

#ifdef LINUX
extern void swab(const void *from, void *to, ssize_t n);
#endif

/* X-DEFINEs ------------------------------------------------------ */

#ifdef FONTSIZE
int PXh=FONTSIZE;
#else
#ifdef MACOSX
int PXh=12;
#else
int PXh=14;
#endif
#endif
#define PXw         (10+(PXh-14)/2)
#define XXh         (PXh+4)
 
enum file_enum {
            FILE_HELP,
            FILE_D1,
            FILE_EXIT,
            FILE_N,
};
 
enum options_enum {
            OPT_DATAPATH,
            OPT_LOGFILE,
            OPT_N
};

enum guider_enum {
            GDR_FLIP_X,
            GDR_FLIP_Y,
            GDR_D1,
            GDR_HOST,
            GDR_D2,
            GDR_RESET,
            GDR_N
};
 
/* --- */

#define MAX_NGUIDERS 3

/* TYPEDEFs ------------------------------------------------------- */

typedef struct gcam_header_tag {
  int  dimx,dimy,seqn,flag;
  char reserved[24];
} Header;

/* GLOBALs -------------------------------------------------------- */

Application *app;
char logfile[512];
int legacy=1;           // todo remove ?Shec
int vertical=0;                        /* horizontal mode NEW v0.224 */
int gWIDE,gHIGH,eWIDE,eHIGH,pHIGH=101; /* NEW v0.224 */

/* function prototype(s) */


/* STATICs -------------------------------------------------------- */

volatile Bool     done=True;

static MainWindow mwin;                /* main window */
static char       fontname[128];
static int        n_msg=0;
static float      throttle=5.0f;
static int        lock_fd;
static time_t     lock_last;

static Menu       fimenu;              /* file dropdown menu */ 
static Menu       opmenu;              /* options dropdown menu */
static EditWindow utbox,rtbox;

static Guider *guiders[MAX_NGUIDERS];
static int n_guiders=0;

static time_t next_disk=0;
static time_t startup_time;
static int    pa_interval=30;
static float  elev=90,para=0;

static int   baseD=2048,baseB=1,baseI=512;  /* todo default=highRes */
static float pscale=0.023f;

#ifdef SIM_ONLY
static const Bool camstat=False;
#else
static const Bool camstat=True;
#endif
static const Bool require_control_key=True; // IDEA allow False

static char    setup_rc[512];          /* dot-file in $HOME */

static Guider  *focusGuider=NULL;

static pthread_mutex_t scanMutex,mesgMutex;

typedef struct {
  char host[128];
  int  port;
  char serialNumber[32];
  int  num;
} AndorIdentity;
static AndorIdentity ident[MAX_NGUIDERS];
static const char *knownHosts[]={"localhost","minnow",
#ifndef SIM_ONLY
                                 "nuc01","nuc02",
                                 "nuc03",
//                               "minnow01","minnow02","minnow03",
#endif
                                 ""};

/* function prototype(s) ------------------------------------------ */

static void    redraw_text       (void);
static void    redraw_gwin       (Guider*);
static void    redraw_compass    (Guider*);

static void    update_ut         (void);

static void    cb_file           (void*);
static void    cb_options        (void*);

static int     set_exptime       (Guider*,float);
static int     set_pa            (Guider*,double,int);
static void    set_sens          (Guider*,double);
static void    set_mm            (Guider*,int);

static int     do_start          (Guider*,int);
static int     do_stop           (Guider*,int);

static void    cb_guider         (void*);

static void    handle_command    (Guider*,const char*);
static int     handle_key        (void*,XEvent*); 
static int     handle_msmode     (void*,XEvent*);

static void*   run_scan          (void*);
static void*   run_housekeeping  (void*);
static void*   run_wait          (void*);
static void*   run_init          (void*);
static void*   run_setup         (void*);
static void*   run_cycle         (void*);
static void*   run_tele          (void*);
static void*   run_display       (void*);
static void*   run_guider        (void*);
static void*   run_write         (void*);
static void*   run_tcpip         (void*); // TODO 1 per guider or gui ?

static int     check_datapath    (char*,int);
static void    my_shutdown       (Bool);

/* ---------------------------------------------------------------- */
 
static void swap4(void* ptr,int nitems)
{
  register int  i;
  register char c,*p=(char*)ptr;

  assert(sizeof(int) == 4);

  for (i=0; i<nitems; i++,p+=4) {
    c = p[0]; p[0] = p[3]; p[3] = c;
    c = p[1]; p[1] = p[2]; p[2] = c;
  }
}

/* ---------------------------------------------------------------- */

static void rotate(double xi,double yi,double pa,double* az,double* el)
{
  double radpa = fabs(pa) / RADS;
  double sinpa = sin(radpa);
  double cospa = cos(radpa);
  assert(radpa != 0);
  if (pa < 0) {                        /* rotate into AZ,EL coodinates */
    *az = +(xi * cospa) -(yi * sinpa);
    *el = -(xi * sinpa) -(yi * cospa);
  } else {
    *az = -(xi * cospa) -(yi * sinpa);
    *el = +(xi * sinpa) -(yi * cospa);
  }
}

/* ---------------------------------------------------------------- */

/* --- M A I N ---------------------------------------------------- */

int main(int argc,char **argv)
{
  int        i,j,ng=0,ni=0;
  int        sleeptime=0,winpos=CBX_TOPLEFT;
  int        x,y,w,h,d,tmode=0,err;
  char       buffer[1024],buf[128];
  char       xserver[256];
  XEvent     event;
  XSizeHints hint;

  /* here comes the code ------------------------------------------ */ 

  lock_fd = lockfile_open(LOCKFILE);    /* open lock file v0066 */
  if (lock_fd == -1) {                  /* failed */
    fprintf(stderr,"%s: cannot open lock-file '%s'\n",P_TITLE,LOCKFILE);
    exit(1);
  } else if (lock_fd == -2) {
    fprintf(stdout,"%s: cannot lock '%s'\n",P_TITLE,LOCKFILE);
    fprintf(stdout,"\tanother instance of '%s' may be running\n",argv[0]);
    exit(2);
  }
  startup_time = lock_last = cor_time(0);

  strcpy(xserver,genv2("DISPLAY",""));
  strcpy(fontname,genv2("ANDORFONT",DEF_FONT)); 
  mwin.disp = NULL;
#ifdef MACOSX
  winpos = CBX_CalcPosition(45,0);
  baseI = 256;
#endif

  pthread_mutex_init(&scanMutex,NULL);
  pthread_mutex_init(&mesgMutex,NULL);

  for (i=0; i<MAX_NGUIDERS; i++) {     /* clear structures */
    strcpy(ident[i].host,"");
    ident[i].port = 0;
    strcpy(ident[i].serialNumber,"");
    ident[i].num = 0;
  }

  { extern char *optarg; char *p;      /* parse command line */  
    extern int opterr,optopt; opterr=0;
    while ((i=getopt(argc,argv,"f:g:h:i:m:n:t:v")) != EOF) {
      switch (i) {
      case 'f':                        /* fontname */
        strcpy(fontname,optarg);
        break;
      case 'g':                        /* guider number --> filename  */
        if (ng == MAX_NGUIDERS) break;
        ident[ng].num  = imin(3,abs(atoi(optarg)));
        ng++;
        break;
      case 'h':                        /* explicit host */
        if (ni == MAX_NGUIDERS) break;
        strcpy(buffer,optarg);
        p = strchr(buffer,':'); if (p) *p = '\0';
        strcpy(ident[ni].host,buffer);
        ident[ni].port = (p) ? SERVER_PORT+atoi(p+1) : SERVER_PORT; 
        ni++;
        break; 
      case 'i':                        /* identity = serial */
        if (ni == MAX_NGUIDERS) break;
        strcpy(ident[ni].serialNumber,optarg);
        ni++;
        break;
      case 'm':                        /* optical mode */
        if (optarg[0] == 'l') {        /* Legacy mode */
          baseD=2000; baseB=4; baseI=500; legacy=2; pHIGH = 107;
          pscale = 0.092f;
        } else
        if (optarg[0] == 'h') {        /* high-res mode todo default */
          baseD=1200; baseB=1; baseI=600; legacy=1; pHIGH = 102;
          pscale = 0.074;
        } else 
        if ((optarg[0] == '1') || (optarg[0] == 'f')) { // full (default)
          baseD=2048; baseB=1; baseI=512; legacy=1; pHIGH = 101;
          pscale = 0.023f;
        }
        break;
      case 't':                        /* TCS */
        tmode = atoi(optarg);
        break;
      case 'v':                        /*  vertical layout NEW v0224 */
        vertical = 1;
        break;
      case '?':
        fprintf(stderr,"%s: option '-%c' unknown or parameter missing\n",
                P_TITLE,(char)optopt);
        break;
      }
    } /* endwhile(getopt) */
  }
  
  n_guiders = imax(ni,ng);
  if (n_guiders == 0) {
    fprintf(stderr,"%s: no guiders (hosts) defined\n",argv[0]);
    lockfile_close(lock_fd);
    unlink(LOCKFILE);
    exit(1);
  }
  for (i=0; i<n_guiders; i++) {
    if (ident[i].num == 0) {           /* no guider number defined */
      if (n_guiders == 1) ident[i].num = MODE_SV;
      else                ident[i].num = 1+i;
    }
  }
#if (DEBUG > 0)
  for (i=0; i<n_guiders; i++) {
    printf("%d: h=%s p=%d i=%s g=%d\n",i,ident[i].host,ident[i].port,
           ident[i].serialNumber,ident[i].num);
  }
#endif

  if (vertical) {                      /* NEW v0224 */
    gWIDE = 632;
    eWIDE = 6+(n_guiders*(2+gWIDE));
#ifdef MACOSX
    eHIGH = 810;
#else
    eHIGH = 1144;                      /* MATE desktop @1920x1200 */
#endif
  } else {
    assert(n_guiders == 1);
    gWIDE = 632+4+baseI; 
    eWIDE = 6+1*(2+632)+baseI+4;
    eHIGH = baseI+33;
  }
  gHIGH = eHIGH-XXh-9;

  InitRandom(0,0,0);                   /* QlTool uses Random */

  /* -------------------------------------------------------------- */

  switch (tmode) {
  case 0:                              /* offline */
    err = telio_init(TELIO_NONE,0);  
    break;
  case 1:                              /* Baade */
    err = telio_init(TELIO_MAG1,5801); 
    break;
  case 2:                              /* Clay */
    err = telio_init(TELIO_MAG2,5801); 
    break;
  default:                             /* Simulator */
    err = telio_init(TELIO_NONE,5801); 
    break;
  }
  if (err) {
    fprintf(stderr,"%s: failed to connect to TCSIS (err=%d)\n",P_TITLE,err);
    lockfile_close(lock_fd);           /* v0067 */
    unlink(LOCKFILE);
    exit(3);
  }

  /* initialize X-stuff ------------------------------------------- */

  Status s = XInitThreads();
  assert(s);

  if ((app=CBX_InitXlib(xserver)) == NULL) {  /* connect to 'xserver' */
    fprintf(stderr,"%s: CBX_InitXlib(%s) failed\n",P_TITLE,xserver);
    lockfile_close(lock_fd);           /* v0067 */
    unlink(LOCKFILE);
    exit(4);
  }
 
  /* get setup from files */

  log_path(logfile,"ANDORLOG","andorgui"); /* create logfile name */
  lnk_logfile(logfile,"andorgui.log");     /* link to $HOME */
  sprintf(buf,"AndorGUI-v%s",P_VERSION);
  message(NULL,buf,MSG_FILE);

  (void)set_path(setup_rc,"andorguirc");   /* $HOME */

  get_string(setup_rc,DBE_DATAPATH,buffer,genv2("HOME","/tmp"));
  i = check_datapath(buffer,0);
  if (!i) strcpy(buffer,genv2("HOME","/tmp"));
  // printf("path= %s\n",buffer);

  for (i=0; i<n_guiders; i++) {
    guiders[i] = (Guider*)malloc(sizeof(Guider)); 
    Guider *g = guiders[i];
    g->num = ident[i].num;
    sprintf(g->name,"Guider%d",g->num);
    if (ident[i].port) {               /* explicit host */
      g->andor = andor_create(ident[i].host,ident[i].port);
    } else {
      g->andor = NULL;
    }
    g->gid = 0;                        /* guiding thread ID */
    g->loop_running = False; 
    g->stop_flag = False;
    g->stabilized = False;
    g->init_flag = g->send_flag = g->write_flag = 0;
    pthread_mutex_init(&g->mutex,NULL);
    g->fps = g->flux = g->ppix = g->back = g->fwhm = g->dx = g->dy = 0;
    g->sens = g->pa = 0.0;
    g->shmode = (g->num == MODE_SH) ? 1 : 0;
    g->pamode = 1;                     /* v0066 */
    g->esmode = 0;
    g->msmode = 1;
    g->px = pscale;
    strcpy(g->send_host,telio_host);   /* v0066 */
    g->send_port = 5700+g->num-1;
    FITSpars *st = &g->status;
    sprintf(buf,FMT_RUN,g->num);
    st->run = (int)get_long(setup_rc,buf,1);
    st->seqNumber = 0;
    get_string(setup_rc,DBE_PREFIX,buf,DEF_PREFIX);
    sprintf(st->prefix,"%s%d_",buf,g->num);
    strcpy(st->datapath,buffer); 
    strcpy(st->origin,"LCO/OCIW");
    strcpy(st->version,P_VERSION);
    strcpy(st->instrument,"Andor Zyla 4.2+ sCMOS"); // IDEA from Andor?
    st->dimx = st->dimy = baseD;
    st->binning = baseB;
    st->exptime = 0.5f;
    st->pixscale = g->px;              /* FITS header */
  }

  /* create main-window ------------------------------------------- */
 
  sprintf(buf,"%s (v%s)",guiders[0]->status.instrument,P_VERSION);
  CXT_OpenMainWindow(&mwin,winpos,eWIDE,eHIGH,&hint,buf,P_TITLE,True);
  // XSynchronize(mwin.disp,True);
  CBX_SelectInput(&mwin,ExposureMask | KeyPressMask | EnterWindowMask);
  (void)CBX_CreateGfC(&mwin,fontname,"bold",PXh);

  (void)CBX_CreateAutoDrop(&mwin,&fimenu,2,2,5*PXw,XXh,"File",cb_file);
    (void)CBX_AddMenuEntry(&fimenu,"Help",    0);
    (void)CBX_AddMenuEntry(&fimenu,NULL,      CBX_SEPARATOR);
    (void)CBX_AddMenuEntry(&fimenu,"Exit",    0);
 
  (void)CBX_CreateAutoDrop(&mwin,&opmenu,fimenu.x+fimenu.w+PXw/2,fimenu.y,
                           8*PXw,XXh,"Options",cb_options);
    (void)CBX_AddMenuEntry(&opmenu,"DataPath",0); // OPT_DATAPATH
    (void)CBX_AddMenuEntry(&opmenu,"Logfile",0);  // OPT_LOGFILE

  w = 11*PXw;
  x = eWIDE - w - PXw/2;
  y = fimenu.y;
  CBX_CreateAutoOutput(&mwin,&utbox,x,y,w,XXh,"UT");
  x -= w + PXw;
  CBX_CreateAutoOutput(&mwin,&rtbox,x,y,w,XXh,"et");

  for (i=0; i<n_guiders; i++) {        /* guider controls */
    Guider *g = guiders[i];
    g->win = CBX_CreateSimpleWindow(&mwin,2+i*(2+gWIDE),2+XXh+3,
                                    gWIDE,gHIGH,app->grey);
    CBX_SelectInput_Ext(mwin.disp,g->win,ExposureMask | KeyPressMask | 
                        EnterWindowMask );

    /* dropdown menu ---------------------------------------------- */
    CBX_CreateAutoDrop(&mwin,&g->gdmenu,opmenu.x+(1+i)*(opmenu.w+PXw/2),
                       fimenu.y,8*PXw,XXh,g->name,cb_guider);
      CBX_AddMenuEntry(&g->gdmenu,"Flip-X",0);   // GDR_FLIP_X
      CBX_AddMenuEntry(&g->gdmenu,"Flip-Y",0);   // GDR_FLIP_Y
      CBX_AddMenuEntry(&g->gdmenu,NULL,CBX_SEPARATOR);
      CBX_AddMenuEntry(&g->gdmenu,"SendHost",0); // GDR_HOST
      CBX_AddMenuEntry(&g->gdmenu,NULL,CBX_SEPARATOR);
      CBX_AddMenuEntry(&g->gdmenu,"Reset",0);    // GDR_RESET
    /* quicklook tool --------------------------------------------- */
    x = 1;
    y = 1;
    g->qltool = qltool_create(&mwin,g->win,fontname,x,y,
                              g->status.dimx,g->status.dimy,baseI);
    sprintf(g->qltool->name,"QlTool%d",1+i);
    // IDEA: set vrad depending on scale
    /* 1st column ------------------------------------------------- */
    Window p = g->win;
    x += g->qltool->lWIDE + PXw/2;     
    y  = 2;
    w  = 10*PXw;
    CBX_CreateAutoOutput_Ext(&mwin,&g->tcbox,p,x,y,w,XXh,"tc 0");
    y += 3+XXh+PXh/3;                  /* Note: 3+XXh in qltool.c */
    CBX_CreateAutoOutput_Ext(&mwin,&g->mxbox,p,x,y,w,XXh,"mx 0");
    y += 3+XXh+PXh/3;
    CBX_CreateAutoOutput_Ext(&mwin,&g->bkbox,p,x,y,w,XXh,"bk 0");
    y += 3+XXh+PXh/3;
    CBX_CreateAutoOutput_Ext(&mwin,&g->fwbox,p,x,y,w,XXh,"fw 0");
    y += 3+XXh+PXh/3;
    CBX_CreateAutoOutput_Ext(&mwin,&g->dxbox,p,x,y,w,XXh,"dx 0");
    y += 3+XXh+PXh/3;
    CBX_CreateAutoOutput_Ext(&mwin,&g->dybox,p,x,y,w,XXh,"dy 0");
    /* 2nd column ------------------------------------------------- */
    x += g->tcbox.w + PXw/2;           
    y  = 2;
    w  = 8*PXw;
    CBX_CreateAutoOutput_Ext(&mwin,&g->gdbox,p,x,y,w,XXh,"gd  off");
    y += 3+XXh+PXh/3;                  /* Note: 3+XXh in qltool.c */
    CBX_CreateAutoOutput_Ext(&mwin,&g->snbox,p,x,y,w,XXh,"sn    0");
    y += 3+XXh+PXh/3; 
    sprintf(buf,"px  %03d",(int)my_round(1000.0*g->px,0));
    CBX_CreateAutoOutput_Ext(&mwin,&g->pxbox,p,x,y,w,XXh,buf);
    y += 3+XXh+PXh/3;
    sprintf(buf,"bx %4d",1+2*g->qltool->vrad);
    CBX_CreateAutoOutput_Ext(&mwin,&g->bxbox,p,x,y,w,XXh,buf);
    y += 3+XXh+PXh/3;
#ifdef PA_BOX
    d  = 3*PXw;
    CBX_CreateAutoOutput_Ext(&mwin,&g->pabox,p,x+d,y,w-d,XXh,"    0");
#else
    CBX_CreateAutoOutput_Ext(&mwin,&g->pabox,p,x,y,w,XXh,"pa    0");
#endif
    y += 3+XXh+PXh/3;
    CBX_CreateAutoOutput_Ext(&mwin,&g->avbox,p,x,y,w,XXh,"av    0");
    /* 3rd column ------------------------------------------------- */
    x += g->gdbox.w + 3*PXw;
    y  = 2;
    w  = 8*PXw;
    CBX_CreateAutoOutput_Ext(&mwin,&g->gmbox,p,x,y,w,XXh,"gm    3");
    y += 3+XXh+PXh/3;                  /* Note: 3+XXh in qltool.c */
    CBX_CreateAutoOutput_Ext(&mwin,&g->gabox,p,x,y,w,XXh,"ga    1");
    y += 3+XXh+PXh/3;
    CBX_CreateAutoOutput_Ext(&mwin,&g->fmbox,p,x,y,w,XXh,"fm    1");
    y += 3+XXh+PXh/3;
    CBX_CreateAutoOutput_Ext(&mwin,&g->mmbox,p,x,y,w,XXh,"mm    0");
    /* 4th column ------------------------------------------------- */
    w  = 15*PXw;
    x = (!vertical) ? 632 - w - PXw/2 : gWIDE - w - PXw/2;
    y = 2;
    CBX_CreateAutoOutput_Ext(&mwin,&g->csbox,p,x,y,w,XXh,"cursorStep=1.0");
    /* graphs ----------------------------------------------------- */
    x = 1;
    y = imax(4+g->qltool->lHIGH,2+6*(1+XXh+PXh/3));
    w = (!vertical) ? (632-x-1*PXw)/2 : (gWIDE-x-1*PXw)/2;
    h = pHIGH;
    d = w/2;
    g->g_tc = graph_create(&mwin,g->win,fontname,"tc",x,y,w,h,1,d);
    graph_scale(g->g_tc,0,10000,0);
    x = (!vertical) ? 632 -w - PXw/2 : gWIDE -w - PXw/2;
    g->g_fw = graph_create(&mwin,g->win,fontname,"fw",x,y,w,h,1,d);
    graph_scale(g->g_fw,0.0,2.0,0);
    y += h + 3;
    x = 1;
#ifdef TEST_GUIDER
    g->g_az = graph_create(&mwin,g->win,fontname,"AZ",x,y,w,h,2,d);
#else
    g->g_az = graph_create(&mwin,g->win,fontname,"AZ",x,y,w,h,1,d);
#endif
    graph_scale(g->g_az,-1.0,1.0,1);
    x = (!vertical) ? 632 -w - PXw/2 : gWIDE -w - PXw/2;
#ifdef TEST_GUIDER
    g->g_el = graph_create(&mwin,g->win,fontname,"EL",x,y,w,h,2,d);
#else
    g->g_el = graph_create(&mwin,g->win,fontname,"EL",x,y,w,h,1,d);
#endif
    graph_scale(g->g_el,-1.0,1.0,1);
    /* exposure controls ------------------------------------------ */
    if (vertical) y = g->qltool->iy + g->qltool->iHIGH + 4;
    else          y = y + pHIGH + 4;
    sprintf(buf,"%.2f",g->status.exptime);
    CBX_CreateAutoOutput_Ext(&mwin,&g->tfbox,g->win,1+2*PXw,y,5*PXw,XXh,buf);
    CBX_CreateAutoOutput_Ext(&mwin,&g->dtbox,g->win,
                             g->tfbox.x+g->tfbox.w+3*PXw/2, y,3*PXw,XXh,"0");
    CBX_CreateAutoOutput_Ext(&mwin,&g->sqbox,g->win,
                             g->dtbox.x+g->dtbox.w+3*PXw/2, y,7*PXw+2,XXh,"0");
    CBX_CreateAutoOutput_Ext(&mwin,&g->fpbox,g->win,
                             g->sqbox.x+g->sqbox.w+4*PXw,   y,5*PXw,XXh,"0");
    CBX_CreateAutoOutput_Ext(&mwin,&g->fdbox,g->win,
                             g->fpbox.x+g->fpbox.w+3*PXw/2, y,5*PXw,XXh,"0");
    CBX_CreateAutoOutput_Ext(&mwin,&g->fgbox,g->win,
                             g->fdbox.x+g->fdbox.w+3*PXw/2, y,5*PXw,XXh,"0");
    y += XXh+PXh/3;
    w  = g->fpbox.x+g->fpbox.w-1;
    CBX_CreateAutoOutput_Ext(&mwin,&g->cmbox,g->win,1,y,w,XXh,"_");
    CBX_CreateAutoOutput_Ext(&mwin,&g->tmbox,g->win,
                             g->fgbox.x,y,g->fgbox.w,XXh,"");
    x  = 1;
    w  = g->fgbox.x-1+5*PXw;
    y += XXh+PXh/3;
    n_msg = (gHIGH-y-3)/XXh;
    g->msbox = (EditWindow*)malloc(n_msg*sizeof(EditWindow));
    g->led   = (Led*)malloc(n_msg*sizeof(Led));
    d = PXw+6;
    for (j=0; j<n_msg; j++) {
      CBX_CreateAutoOutput_Ext(&mwin,&g->msbox[j],g->win,x+d,y,w-d,XXh,NULL);
      CBX_CreateLed_Ext(&mwin,&g->led[j],g->win,x+2,y+4,PXw,PXw,app->lgrey); 
      y += XXh;
    }
  } /* endfor(n_guiders) */
  done = False;

  message(NULL,guiders[0]->status.instrument,MSG_FILE);

  /* -------------------------------------------------------------- */

  for (i=0; i<n_guiders; i++) {        /* setup GUI */
    // redraw_gwin(guiders[i]);
    // CBX_SendExpose_Ext(mwin.disp,guiders[i]->win);
    set_mm  (guiders[i],3);            /* mouse mode */
    set_sens(guiders[i],0.5);          /* guider sensitivity */
    set_pa  (guiders[i],360.0,0);
    if (guiders[i]->andor) thread_detach(run_init,guiders[i]);
  }

  thread_detach(run_scan,NULL);

  /* this loop drives the program --------------------------------- */

  while (!done) {                      /* main loop */
    while (!CBX_AutoEvent(&mwin,&event,&sleeptime)) { /* no event */
      CBX_Sleep(&sleeptime,5,350);
      update_ut();
      if (done) break;                 /* break main loop */
    }
    if (done) break;                   /* break main loop */
#if (DEBUG > 3)
    fprintf(stderr,"%s: event.type=%d\n",__FILE__,event.type);
#endif
    switch(event.type) {
    case Expose:                       /* main window exposed  */
      if (event.xexpose.count == 0) {  /* if last event in queue */
        if (event.xexpose.window == mwin.win) {
          redraw_text();               /* redraw everything */
          break;
        }
        for (i=0; i<n_guiders; i++) {
          Guider *g = guiders[i];
          if (event.xexpose.window == g->win) { 
            redraw_compass(g);
            // implicit redraw_gwin(g); 
            break; 
          }
          if (qltool_event(g->qltool,&event)) break;
          if (graph_event(g->g_tc,&event)) break;
          if (graph_event(g->g_fw,&event)) break;
          if (graph_event(g->g_az,&event)) break;
          if (graph_event(g->g_el,&event)) break;
        }
      }
      break;
    case MapNotify:                    /* window mapped */
    case UnmapNotify:                  /* window unmapped */
#if (DEBUG > 0)
      printf("%s: Un/Map Notify\n",__FILE__);
#endif
      break;
    case EnterNotify:                  /* enter window */
      if (event.xany.window == mwin.win) {
        if (focusGuider) {
          //CBX_WindowBorder_Ext(mwin.disp,focusGuider->win,app->black);
          CBX_WindowBorder_Ext(mwin.disp,focusGuider->cmbox.win,app->black);
          focusGuider = NULL; 
        }
      } else {
        for (i=0; i<n_guiders; i++) {
          Guider *g = guiders[i];
          if (event.xany.window == g->win) {
            if (g != focusGuider) {
             if (focusGuider) {        /* green border ?Shec */
              // CBX_WindowBorder_Ext(mwin.disp,focusGuider->win,app->black);
              CBX_WindowBorder_Ext(mwin.disp,focusGuider->cmbox.win,app->black);
             }
             focusGuider = g;
             // CBX_WindowBorder_Ext(mwin.disp,g->win,app->green);
             CBX_WindowBorder_Ext(mwin.disp,g->cmbox.win,app->green);
            }
            break;
          }
        }
      }
      break;
    case DestroyNotify:                /* WM closed window */
#if (DEBUG > 0)
      printf("%s: DestroyNotify on window=%d\n",__FILE__,
             (int)event.xdestroywindow.window);
#endif
      break;
    case MotionNotify:                 /* Mouse */
#if (DEBUG > 0)
      printf("%s: Motion Notify\n",__FILE__);
#endif
      /* ignore */
      break;
    case ReparentNotify:               /* MacOS */
#if (DEBUG > 0)
      printf("%s: Reparent Notify\n",__FILE__);
#endif
      /* ignore */
      break;
    case ConfigureNotify:              /* window moved */
#if (DEBUG > 0)
      printf("%s: Configure Notify\n",__FILE__);
#endif
      /* ignore */
      break;
    case KeyPress:                     /* keyboard events */
      if (event.xkey.window == mwin.win) {
#ifdef SIM_ONLY                        /*  */
        if (CBX_GetKey(&event) == 'Q') {
          my_shutdown(True);
        }
#endif
      } else {
        for (i=0; i<n_guiders; i++) {
          Guider *g = guiders[i];
          if (event.xkey.window == g->win) {
            if (qltool_handle_key(g->qltool,(XKeyEvent*)&event)) break;
            if (handle_key(g,&event)) break;
          } 
        }
      }
      break;
    case ButtonPress:                  /* mouse button pressed */
      if (event.xbutton.button == 1) { /* left button */
        for (i=0; i<n_guiders; i++) {
          Guider *g = guiders[i];
          if (g->init_flag != 1) continue;
          if (event.xbutton.window == g->qltool->iwin) {
            if (require_control_key) {
              if (!(event.xbutton.state & ControlMask)) {
                message(g,"Control key required",MSG_WARN);
                break; 
              }
            }
            handle_msmode(g,&event);
            break;
          }
        } /* endfor(guiders) */
      } /* endif(left button) */
      break;
#if (DEBUG > 0)
    default:
      fprintf(stderr,"unknown event '%d'\n",event.type);
      break;
#endif
    } /* endswitch(event.type) */
  } /* while(!done) */                 /* end of main-loop */

  /* free up all the resources and exit --------------------------- */

#if 0 // TODO
  if (info.port != 0) {                /* TCP/IP enabled */
    (void)TCPIP_TerminateServerThread(&info,1000);
  }
#endif

  CBX_CloseMainWindow(&mwin);

  for (i=0; i<n_guiders; i++) {
    Guider *g = guiders[i];
    if (g->andor) andor_free(g->andor);
  }

#if (TIME_TEST > 0)                    /* print cpu-time */
  { struct tms tmsbuf;
    (void)times(&tmsbuf);
    printf("%s(%d): total-cpu: %5.2f seconds\n",P_TITLE,(int)getpid(),
      (float)(tmsbuf.tms_stime)/(float)sysconf(_SC_CLK_TCK) +
      (float)(tmsbuf.tms_utime)/(float)sysconf(_SC_CLK_TCK));
  }
#endif
 
  CXT_CloseLib(True);
  lockfile_close(lock_fd);
  unlink(LOCKFILE);
 
  return 0;
}

/* ---------------------------------------------------------------- */

static void draw_compass(Guider* g,int x,int y,int r,double north,double east,
                         u_long color)
{
  Display *disp=mwin.disp;
  Window  win=g->win;
  GC      gc=mwin.gfs.gc;
  int     x2,y2;

  CBX_Lock(0);
  XDrawArc(disp,win,gc,x-r,y-r,2*r,2*r,0,23040);
  XSetLineAttributes(disp,gc,2,LineSolid,CapRound,JoinRound);
  XSetForeground(disp,gc,color);
  x2 = x + (int)floor(r*sin(north/RADS)+0.5);
  y2 = y - (int)floor(r*cos(north/RADS)+0.5);
  XDrawLine(disp,win,gc,x,y,x2,y2);
  x2 = x + (int)floor(0.5*r*sin(east/RADS)+0.5);
  y2 = y - (int)floor(0.5*r*cos(east/RADS)+0.5);
  XDrawLine(disp,win,gc,x,y,x2,y2);
  XSetLineAttributes(disp,gc,1,LineSolid,CapRound,JoinRound);
  XSetForeground(disp,gc,app->black);
  CBX_Unlock();
}

/* --- */

static void redraw_compass(Guider* g)
{
  int    x,y,r=XXh;
  double north,east;
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p): pa=%.1f, para=%.1f\n",PREFUN,g,g->pa,para);
#endif
  assert(g->pa != 0);

  CBX_Lock(0);
  XClearWindow(mwin.disp,g->win);
  CBX_Unlock();

  x = g->gmbox.x + r;
  y = g->dybox.y - PXh/3;
  north = -fabs(g->pa) + para;         /* N/E */
  east  = north + ((g->pa > 0) ? 90.0 : -90.0);
  draw_compass(g,x,y,r,north,east,app->green); 

  x += 2*r + PXw;
  north = -fabs(g->pa);                /* az,el */
  east  = north + ((g->pa < 0) ? 90.0 : -90.0);
  draw_compass(g,x,y,r,north,east,app->red); 

  redraw_gwin(g);
}

/* ---------------------------------------------------------------- */

static void redraw_gwin(Guider *g)
{
  int     j,x,y;
  char    buf[128];
  Display *disp=mwin.disp;
  Window  win=g->win;
  GC      gc=mwin.gfs.gc;
#if (DEBUG > 1)
  fprintf(stderr,"%s(%s)\n",PREFUN,g->name);
#endif

  (void)CBX_Lock(0);
  XDrawString(disp,win,gc,g->tfbox.x-2*PXw,g->tfbox.y+PXh,"tf",2);
  XDrawString(disp,win,gc,g->fpbox.x-3*PXw,g->fpbox.y+PXh,"FpS",3);
  XDrawString(disp,win,gc,g->tmbox.x-4*PXw,g->tmbox.y+PXh,"Temp",4);
  for (j=0; j<QLT_NCURSORS; j++) {
    if (g->qltool->cursor_mode == j) XSetForeground(disp,gc,app->green);
    x = g->qltool->cubox[j].x-PXw;
    y = g->qltool->cubox[j].y+PXh;
    sprintf(buf,"%d",1+j);
    XDrawString(disp,win,gc,x,y,buf,strlen(buf));
    if (g->qltool->cursor_mode == j) XSetForeground(disp,gc,app->black);
  }
#ifdef PA_BOX
  if (g->pamode) XSetForeground(disp,gc,app->green);
  XDrawString(disp,win,gc,g->pabox.x-2*PXw-PXw/2,g->pabox.y+PXh,"pa",2);
  if (g->pamode) XSetForeground(disp,gc,app->black);
#endif
  CBX_Unlock();
  sprintf(g->csbox.text,"cursorStep=%.1f",g->qltool->cursor_step);
  CBX_UpdateEditWindow(&g->csbox);
}

/* --- */

static void redraw_text(void)
{
#if 0 // not used
  Display *disp=mwin.disp;
  Window  win=mwin.win;
  GC      gc=mwin.gfs.gc;
#endif
#if (DEBUG > 2)
  fprintf(stderr,"%s: %s()\n",__FILE__,PREFUN);
#endif

  (void)CBX_Lock(0);

  CBX_Unlock();
}

/* ---------------------------------------------------------------- */

static void update_ut(void)
{
  char buf[128];
  static time_t ut_now=0,et_next=0,pa_next=0;
#if (DEBUG > 2)
  fprintf(stderr,"%s()\n",PREFUN);
#endif

  if (cor_time(0) == ut_now) return;   /* current time IDEA TCS offtime */
  ut_now = cor_time(0);                /* IDEA depends on time server */

  sprintf(utbox.text,"UT %s",get_ut_timestr(buf,ut_now));
  CBX_UpdateEditWindow(&utbox); 

  if (!camstat) { Guider *g=guiders[0]; /* Simulator */
    if ((ut_now % 2) == 0) { char title[128];
      if ((ut_now % 4) == 0) {         /* blink title */
        sprintf(title,"%s (v%s)",g->status.instrument,P_VERSION);
      } else {
        sprintf(title,"%s - Simulator",g->status.instrument);
#if (__SIZEOF_POINTER__ == 8)          /* __x86_64__ , __LP64__ */
        strcat(title,"/64");
#endif
      }
      CBX_SetMainWindowName(&mwin,title);
    }
  }

  if (ut_now >= lock_last+6000) {      /* refresh lockfile */
    lockfile_close(lock_fd); 
    lock_fd = lockfile_open(LOCKFILE);
    if (lock_fd < 0) { char buffer[1024];
      sprintf(buffer,"re-locking '%s' failed, errno=%d",LOCKFILE,errno);
      tdebug(P_TITLE,buffer);
    }
    lock_last = ut_now;
  }

  if (ut_now >= et_next) {             /* update runtime */
    double runtime = (double)(ut_now - startup_time);
    strcpy(buf,"s"); et_next = ut_now+1;
    if (runtime > 120) {               /* 120 seconds */
      runtime /= 60.0; strcpy(buf,"m"); et_next += 5; /* 0.1 mins */
      if (runtime > 100)  {            /* 100 minutes */
        runtime /= 60.0; strcpy(buf,"h"); et_next += 354; /* 0.1 hrs */
        if (runtime > 96)  { runtime /= 24.0; strcpy(buf,"d"); }
      }
    }
    sprintf(rtbox.text,"et %.1f%s",runtime,buf);
    CBX_UpdateEditWindow(&rtbox); 
    // IDEA return to execute only one thing each time
  }

  if (ut_now >= pa_next) { int i; Guider *g;
    for (i=0; i<n_guiders; i++) {
      g = guiders[i];
      if (g->pamode) {
        thread_detach(run_tele,NULL);
        pa_next = ut_now + pa_interval;
        break;
      }
    }
  }
  
  if (ut_now >= next_disk) { // IDEA thread_detach
    Guider *g = guiders[0];
    float file_system = checkfs(g->status.datapath); 
    if (file_system < 0.05) { char buf[128]; 
      sprintf(buf,"WARNING: disk %.1f%% full",100.0*(1.0-file_system));
      message(g,buf,MSG_WARN);
    }
    next_disk = ut_now+600;
  }
}

/* ---------------------------------------------------------------- */
 
static void cb_file(void* param)
{
  int i;
 
  switch (fimenu.sel) {  
  case FILE_HELP:
    { char cmd[512]; AMB *box;
      box = CBX_AsyncMessageOpen(P_TITLE ": startup browser",DEF_HELP);
      sprintf(cmd,"%s %s/%s &",genv2("BROWSER",DEF_BROWSER),DEF_HELP,MY_HELP);
      system(cmd);
      sleep(3);
      CBX_AsyncMessageClose(box);
    }
    break;
  case FILE_EXIT:                      /* exit GUI */
    done = True;                       /* kill housekeeping loop */
    for (i=0; i<n_guiders; i++) {
      if (guiders[i]->loop_running) {
        if (!CBX_AlertBox("Exposure","Really exit?")) done = False;
        break;
      }
    }
    if (done) my_shutdown(True);
    break;
  }
  CBX_ClearAutoQueue(&mwin);
}

/* ---------------------------------------------------------------- */

static void cb_options(void* param)
{
  int  i;
  char buffer[1024];

  switch (opmenu.sel) {
  case OPT_LOGFILE:                    /* set logfile name */
    CBX_AsyncMessageBox(P_TITLE,logfile);
    break;
  case OPT_DATAPATH:                   /* set save-path (data) */
    sprintf(buffer,"%s",guiders[0]->status.datapath);
    if (CBX_ParameterBox("Set Datapath",buffer)) {
      check_datapath(buffer,1);
      put_string(setup_rc,DBE_DATAPATH,buffer);
      for (i=0; i<n_guiders; i++) {
        strcpy(guiders[i]->status.datapath,buffer);
      }
    }
    break;
  }
  CBX_ClearAutoQueue(&mwin);
}

/* ---------------------------------------------------------------- */

static Guider* get_guider(void* param)
{
  int i;
  Guider *g=NULL;

  for (i=0; i<n_guiders; i++) {
    g = guiders[i];
    if (param == &g->gdmenu) break;
    g = NULL;
  }
  return g;
}

/* --- */

static void cb_guider(void* param)
{
  Menu *m=(Menu*)param;
  Guider *g = get_guider(param);
  char buffer[1024],buf[256];
  assert(g);

  switch (m->sel) {
  case GDR_FLIP_X:
    g->qltool->flip_x = 1-g->qltool->flip_x;
    g->gdmenu.entry[GDR_FLIP_X].flag = (g->qltool->flip_x) ? CBX_CHECKED : 0;
    qltool_redraw(g->qltool,False);
    break;
  case GDR_FLIP_Y:
    g->qltool->flip_y = 1-g->qltool->flip_y;
    g->gdmenu.entry[GDR_FLIP_Y].flag = (g->qltool->flip_y) ? CBX_CHECKED : 0;
    qltool_redraw(g->qltool,False);
    break;
  case GDR_HOST:                       /* v0066 */
    sprintf(buffer,"%s %d",g->send_host,g->send_port);
    if (CBX_ParameterBox("Send Host & Port",buffer)) { int port;
      if (sscanf(buffer,"%s %d",buf,&port) == 2) {
        strcpy(g->send_host,buf);
        g->send_port = port;
      }
    }
    break;
  case GDR_RESET:
    message(g,"not implemented",MSG_WARN);
    break;
  }
}

/* ---------------------------------------------------------------- */

static int do_start(Guider* g,int wait) 
{
  int weDidIt=0;
#if (DEBUG > 0)
  fprintf(stderr,"%s(%s)\n",PREFUN,g->name);
#endif

  if (g->init_flag != 1) return -E_NOTINIT;

  if (!g->loop_running) { 
    message(g,"starting acquisition",MSG_INFO);
    thread_detach(run_cycle,g);
    while (wait > 0) {
      if (!g->loop_running) msleep(350);
      wait -= 350;
    }
    weDidIt = 1;
  }

  return weDidIt;
}

/* ---------------------------------------------------------------------- */

static int do_stop(Guider* g,int wait) 
{
  int weDidIt=0;
#if (DEBUG > 0)
  fprintf(stderr,"%s(%s)\n",PREFUN,g->name);
#endif

  if (g->loop_running) {
    if (g->qltool->guiding) { g->qltool->guiding = 0; msleep(350); } 
    g->stop_flag = True;
    message(g,"stopping acquisition",MSG_INFO);
    while (wait > 0) {
      if (g->loop_running) msleep(350);
      wait -= 350;
    }
    weDidIt = 1;
  }
  return weDidIt;
}

/* ---------------------------------------------------------------------- */

static int handle_msmode(void* param,XEvent* event)
{
  int  err=0;
  char buf[128]="";
  Guider *g = (Guider*)param;

  int    x = event->xbutton.x;
  int    y = event->xbutton.y;
  double r = (event->xbutton.state & Button3Mask) ? 1.0 : 0.1;

  switch (g->msmode) {       /* mouse mode */
  case 1:                    /* move cursor to mouse position */
    sprintf(buf,"box: to mouse position (%.1f)",r);
    qltool_cursor_set(g->qltool,QLT_BOX,x,y,r);
    break;
  case 2: case -2:           /* move probe(s) */
    if (g->msmode > 0) sprintf(buf,"probe-%d offset (%.1f)",g->num,r); 
    else               sprintf(buf,"coordinated offset (%.1f)",r);
    err = telio_open(2); 
    if (!err) {
      if (g->pamode) err = set_pa(g,g->pa,1);
      if (!err) { double dx,dy,da,de;
        qltool_cursor_off(g->qltool,QLT_BOX,x,y,r,&dx,&dy);
        rotate(g->px*dx,g->px*dy,g->pa,&da,&de);
        assert(MODE_SV == 3);          /* 3: move PR&SH */
        err = telio_gpaer(g->num,-da,-de);
        if (!err && (g->msmode < 0)) err = telio_aeg(da,de);
      }
      telio_close();
    }
    break;
  case 3:                    /* move telescope */
    sprintf(buf,"tele: star at mouse to box (%.1f)",r); 
    err = telio_open(2); 
    if (!err) {
      if (g->pamode) err = set_pa(g,g->pa,1);
      if (!err) { double dx,dy,da,de;
        qltool_cursor_off(g->qltool,QLT_BOX,x,y,r,&dx,&dy);
        rotate(g->px*dx,g->px*dy,g->pa,&da,&de);
        err = telio_aeg(da,de);
      }
      telio_close();
    }
    break;
  default:
    fprintf(stderr,"%s: invalid mouse mode %d\n",P_TITLE,g->msmode);
    break;
  }
  if (*buf) message(g,buf,MSG_WINDOW);
  if (err) {
    sprintf(buf,"errorCode= %d",err); message(g,buf,MSG_ERROR);
  }

  return err;
}

/* ---------------------------------------------------------------------- */

static int handle_key(void* param,XEvent* event) 
{
  Guider *g=(Guider*)param;
  int r=1,i;
  static const float steps[]={0,0.1f,1.0f,4.0f,40.0f,0};
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p,%p)\n",PREFUN,param,event);
#endif

  CBX_Lock(0);
  char *keys = XKeysymToString(XLookupKeysym((XKeyEvent*)event,0));
  //printf("%s: keys=%s\n",PREFUN,keys);
  CBX_Unlock();

  if (!strcmp(keys,"F1")) {            /* stop guiding "F1" */
    handle_command(g,"fone");
  } else
  if (!strcmp(keys,"F2")) {            /* calculate here */
    handle_command(g,"ftwo");
  } else
  if (!strcmp(keys,"F3")) {            /* guide here */
    handle_command(g,"fthr");
  } else 
  if (!strcmp(keys,"F4")) {            /* center -- calculate  */
    handle_command(g,"ffou");
  } else
  if (!strcmp(keys,"F5")) {            /* center -- guide */
    handle_command(g,"ffiv");
  } else 
  if (!strcmp(keys,"F9")) {
    handle_command(g,"fnin");
  } else 
  if (!strcmp(keys,"F10")) {
    handle_command(g,"ften");
  } else 
  if (!strcmp(keys,"Insert") || !strcmp(keys,"Delete") ||
      !strcmp(keys,"Prior") || !strcmp(keys,"Next")) { /* cursor step */
    for (i=0; i<(sizeof(steps)/sizeof(float)); i++) {
      if (g->qltool->cursor_step == steps[i]) break;
    }
    if (!strcmp(keys,"Insert") || !strcmp(keys,"Prior")) {
      if (steps[i+1] != 0) g->qltool->cursor_step = steps[i+1];
    } else {
      if (steps[i-1] != 0) g->qltool->cursor_step = steps[i-1];
    }
    redraw_gwin(g);
  } else {
    // printf("%s(): keys=%s\n",PREFUN,keys);
    r = 0;
  }
  if (r == 0) {                        /* not a special key */
    if (event->xkey.state & ControlMask) { /* Control key is pressed */
      if (!strcmp(keys,"1") || !strcmp(keys,"2") || !strcmp(keys,"3") ||
          !strcmp(keys,"4") || !strcmp(keys,"5")) {
        g->qltool->cursor_mode = atoi(keys)-1;
        redraw_gwin(g);
        qltool_redraw(g->qltool,False);
      }
      r = 1;
    }
  }
  if (r == 0) {                        /* redirect to command edit */
    if (g->cmbox.cpos > 0) {           /* remove '_' */
      g->cmbox.cpos--; g->cmbox.text[g->cmbox.cpos] = '\0'; 
    }
    int flag = CBX_HandleEditWindow(&g->cmbox,(XKeyEvent*)event);
    if (flag) {                        /* <Enter> or <Return> */
      if (strlen(g->cmbox.text) > 0) {
        handle_command(g,g->cmbox.text);
        strcpy(g->cmbox.text,"");
        CBX_ResetEditWindow(&g->cmbox);
        // IDEA g->cmbox.cpos = 1; strcpy(g->cmbox.text,"_");
      }
    } else {
      g->cmbox.cpos += 1; strcat(g->cmbox.text,"_");
      CBX_UpdateEditWindow(&g->cmbox);
    }
  }

  return r;
}

/* ---------------------------------------------------------------- */

static void handle_command(Guider* g,const char* command)
{
  int  err=0,n=0;
  char cmd[128]="",par1[128]="",par2[128]="",par3[128]="",msg[256]="";
  char buf[256];
#if (DEBUG > 1)
  fprintf(stderr,"%s(%s,%s)\n",PREFUN,g->name,command);
#endif

  //sprintf(buf,"%s: %s",PREFUN,command); message(g,buf,MSG_FLUSH);

  if (strlen(command) < 32) {                 /* v0065 */
    char *p1=(char*)command,*p2=buf;
    do {
      if (isupper(*p1) != islower(*p1)) {
        *p2 = *p1; p1++; p2++;
      } else {
        if (*p1 != ' ') { *p2 = ' '; p2++; }
        strcpy(p2,p1); 
        break;
      }
    } while (*p1);
    *p2 = *p1;
    message(g,buf,MSG_WINDOW);
    n = sscanf(buf,"%s %s %s %s",cmd,par1,par2,par3);
  }

  if (!strcasecmp(cmd,"apa")) {        /* toggle camera position angle */
    if (*par1) {
      g->pamode = 1; pa_interval = imax(5,atoi(par1));
    } else {
      g->pamode = (g->pamode) ? 0 : 1;
    }
    if (g->pamode) set_pa(g,g->pa,1);  /* get posAngle from TCS */
#ifdef PA_BOX
    redraw_gwin(g); 
#else
    g->pabox.fg = (g->pamode) ? app->green : app->black;
    CBX_UpdateEditWindow(&g->pabox);
#endif
  } else
  if (!strncasecmp(cmd,"av",2)) {      /* AVF,AVG */
    if (n >= 2) {
#if 1 // rolling at GUI
      g->andor->rolling = imax(0,imin(99,atoi(par1)));
#else // rolling at server
      err = andor_rolling(g->andor,imax(0,imin(99,atoi(par1))));
#endif
    }
    sprintf(g->avbox.text,"av %4d",g->andor->rolling);
    CBX_UpdateEditWindow(&g->avbox);
  } else 
  if (!strcasecmp(cmd,"bf") || !strcasecmp(cmd,"bg")) {
    err = E_NOTIMP;
  } else 
  if (!strcasecmp(cmd,"bx")) {         // guider box size ?Shec legacy ?
    if (n >= 2) g->qltool->vrad = imax(7,(atoi(par1)-1)/2);
    sprintf(msg,"box=%d",1+2*g->qltool->vrad);
    sprintf(g->bxbox.text,"bx %4d",1+2*g->qltool->vrad);
    CBX_UpdateEditWindow(&g->bxbox);
  } else
  if (!strcasecmp(cmd,"close")) {      /* TESTING */
    if (g->loop_running) do_stop(g,5000);
    andor_close(g->andor);
    g->init_flag = 0;
  } else
  if (!strcasecmp(cmd,"es")) {         /* toggle extended guiding ?Shec */
    g->esmode = (g->esmode) ? 0 : 1;
  } else
  if (!strcasecmp(cmd,"exit")) {       /* exit program */
    my_shutdown(True);
  } else
  if (!strcasecmp(cmd,"fone")) {       /* F1 */
    g->qltool->guiding = 0;
    if (g->gid) { pthread_join(g->gid,NULL); g->gid = 0; }
    sprintf(g->gdbox.text,"gd  off"); CBX_UpdateEditWindow(&g->gdbox);
  } else
  if (!strcasecmp(cmd,"ftwo")) {       /* F2 */
    if (g->loop_running) {             /* implies (init_flag==1) */
      g->qltool->guiding = -2;
      if (!g->gid) thread_create(run_guider,g,&g->gid);
      sprintf(g->gdbox.text,"gd calc"); CBX_UpdateEditWindow(&g->gdbox);
    } else err = E_NOTACQ;
  } else 
  if (!strcasecmp(cmd,"fthr")) {       /* F3 */
    if (g->loop_running) {             /* implies (init_flag==1) */
      g->qltool->guiding = -3;
      if (!g->gid) thread_create(run_guider,g,&g->gid); 
      sprintf(g->gdbox.text,"gd move"); CBX_UpdateEditWindow(&g->gdbox);
    } else err = E_NOTACQ;
  } else
  if (!strcasecmp(cmd,"ffou")) {       /* F4 */
    if (g->loop_running) {             /* implies (init_flag==1) */
      g->qltool->guiding = -4;
      if (!g->gid) thread_create(run_guider,g,&g->gid);
      sprintf(g->gdbox.text,"gd calc"); CBX_UpdateEditWindow(&g->gdbox);
    } else err = E_NOTACQ;
  } else
  if (!strcasecmp(cmd,"ffiv")) {       /* F5 */
    if (g->loop_running) {             /* implies (init_flag==1) */
      g->qltool->guiding = -5;
      if (!g->gid) thread_create(run_guider,g,&g->gid);
      sprintf(g->gdbox.text,"gd move"); CBX_UpdateEditWindow(&g->gdbox);
    } else err = E_NOTACQ;
  } else
  if (!strcasecmp(cmd,"fnin")) {       /* F9 TODO ?Shec,?Povilas */
    printf("send FLDn command to the guider probe control program\n");
  } else
  if (!strcasecmp(cmd,"ften")) {       /* F10 TODO ?Shec,?Povilas */
    printf("send SHAn command to the guider probe control program\n");
  } else
  if (!strcasecmp(cmd,"gm") || !strcasecmp(cmd,"fm")) { 
    err = E_NOTIMP;
  } else
  if (!strcasecmp(cmd,"mc")) {         /* simulate mouse click ?Shec */
    err = E_NOTIMP;
  } else
  if (!strcasecmp(cmd,"mm")) {         /* mouse mode */
    set_mm(g,atoi(par1));
  } else
  if (!strcasecmp(cmd,"pa")) {         /* position angle */
    set_pa(g,atof(par1),0);
  } else
  if (!strcasecmp(cmd,"px")) {         /* pixel scale */
    g->px = fmax(0.001,atof(par1)/1000.0); 
    int px = (int)my_round(1000.0*g->px,0);
    sprintf(g->pxbox.text,(px >= 1000) ? "px %d" : "px  %03d",px);
    CBX_UpdateEditWindow(&g->pxbox);
  } else 
  if (!strcasecmp(cmd,"send")) {       /* send frame via TCP/IP */
    if (g->init_flag == 1) {
      g->send_flag = (n > 1) ? atoi(par1) : -1;
    } else { err = E_ERROR; strcpy(msg,"Andor not initialized"); }
  } else
  if (!strcasecmp(cmd,"sh")) {         /* Shack-Hartman mode */
    if (n > 1) {
       g->shmode = atoi(par1);
#ifdef SIM_ONLY
      sprintf(cmd,"simmode %s",par1);
      andor_sim(g->andor,cmd,NULL);
#endif
    } else err = E_MISSPAR;            /* missing parameter */
  } else
  if (!strcasecmp(cmd,"sky") || !strcasecmp(cmd,"sub")) {
    err = E_NOTIMP;
  } else
  if (!strcasecmp(cmd,"sn")) {         /* guider sensitivity */
    set_sens(g,atof(par1));
  } else
  if (!strcasecmp(cmd,"span") || !strcasecmp(cmd,"zero") ||
      !strcasecmp(cmd,"pct")  || !strcasecmp(cmd,"bkg")) {
    qltool_scale(g->qltool,cmd,par1,par2);
  } else 
  if (!strcasecmp(cmd,"tc")) {         /* send a TCS command */
    err = E_NOTIMP;
    printf("send a TCS command %s %s %s\n",par1,par2,par3); // TODO 
  } else
  if (!strcasecmp(cmd,"tec")) {        /* cooler on/off */
    int weStoppedTheLoop = do_stop(g,5000);
    err = andor_tempcon(g->andor,atoi(par1)); 
    if (weStoppedTheLoop) do_start(g,0);
  } else
  if (!strcasecmp(cmd,"tf") || !strcasecmp(cmd,"tg")) {
    if (n >= 2) set_exptime(g,atof(par1));
  } else
  if (!strcasecmp(cmd,"xy")) {         /* activate cursor 1-5 */
    int j = atoi(par1)-1;
    if ((j >= 0) && (j < QLT_NCURSORS)) { 
      g->qltool->cursor_mode = j;      /* make active */
    }
    redraw_gwin(g);
  } else 
  if (!strcasecmp(cmd,"xys")) {        /* set cursor */
    if (n > 3) {
      int j = imax(0,imin(QLT_NCURSORS-1,atoi(par1)-1));
      g->qltool->curx[j] = (float)atof(par2)/legacy;
      g->qltool->cury[j] = (float)atof(par3)/legacy;
      g->qltool->cursor_mode = j;      /* make active */
    } else 
    if (n > 2) { 
      g->qltool->curx[g->qltool->cursor_mode] = (float)atof(par1)/legacy;
      g->qltool->cury[g->qltool->cursor_mode] = (float)atof(par2)/legacy;
    } else err = E_MISSPAR;
    if (!err) {
      redraw_gwin(g);
      qltool_redraw(g->qltool,False);
    }
  } else
  if (!strcasecmp(cmd,"xyr")) {        /* move cursor 1-5 */
    if (n > 3) {
      int j = imax(0,imin(QLT_NCURSORS-1,atoi(par1)-1));
      g->qltool->curx[j] += (float)atof(par2)/legacy;
      g->qltool->cury[j] += (float)atof(par3)/legacy;
      g->qltool->cursor_mode = j;      /* make active */
    } else 
    if (n > 2) { 
      g->qltool->curx[g->qltool->cursor_mode] += (float)atof(par1)/legacy;
      g->qltool->cury[g->qltool->cursor_mode] += (float)atof(par2)/legacy;
    } else err = E_MISSPAR;
    if (!err) {
      redraw_gwin(g);
      qltool_redraw(g->qltool,False);
    }
  } else
  /* --- extended commmand set ------------------------------------ */
  if (!strncasecmp(cmd,"lmag",3)) {    /* lupe magnification */
    qltool_lmag(g->qltool,atoi(par1));
    sprintf(msg,"lmag=%d",g->qltool->lmag);
  } else
  if (!strncasecmp(cmd,"lut",3)) {     /* color lookup table */
    if (n >= 2) qltool_lut(g->qltool,par1);
    else        err = E_MISSPAR;
  } else
  if (!strncasecmp(cmd,"scale",3)) {   /* scaling mode */
    qltool_scale(g->qltool,par1,par2,par3);
  } else
  if (!strncasecmp(cmd,"dt",2)) {      /* display throttle */
    if (*par1) throttle = fmax(0,atof(par1));
    else               sprintf(msg,"dt= %.0f",throttle);
  } else
  if (!strncasecmp(cmd,"smooth",2)) {  /* IDEA display on GUI */
    g->qltool->smoothing = abs(atoi(par1));
  } else 
  if (!strncasecmp(cmd,"start",4)) {   /* start exposure loop */
    int r = do_start(g,0);
    if (r < 0) err = -r;
  } else
  if (!strncasecmp(cmd,"stop",3)) {    /* stop exposure loop */
    do_stop(g,0);
  } else
  if (!strncasecmp(cmd,"write",3)) {   /* write files */
    g->write_flag = (n > 1) ? atoi(par1) : 1;
    thread_detach(run_write,g);
  } else
  if (!strncasecmp(cmd,"sim",3)) {     /* simulator commands */
    err = andor_sim(g->andor,command,msg);
    if (*par1) strcpy(msg,"");
  } else                               /* reset andorserver */
  if (!strcasecmp(cmd,"dspi") || !strcasecmp(cmd,"reset")) {
    if (g->init_flag >= 0) thread_detach(run_init,g);
#ifdef SIM_ONLY
  } else                               /* restart andorserver */
  if (!strncasecmp(cmd,"restart",4)) {
    if (g->loop_running) do_stop(g,3000);
    err = andor_server(g->andor,cmd,buf);
    if (!err) thread_detach(run_wait,g);
#endif
  } else                               /* reboot Minnow */
  if (!strncasecmp(cmd,"reboot",4)) {
    if (g->loop_running) do_stop(g,3000);
    err = andor_server(g->andor,cmd,buf);
    if (!err) thread_detach(run_wait,g);
  } else                               /* shutdown Minnow */
  if (!strncasecmp(cmd,"shutdown",4) || !strncasecmp(cmd,"poweroff",5)) {
    if (g->loop_running) do_stop(g,3000);
    err = andor_server(g->andor,cmd,buf);
  } else {
    err = E_ERROR; sprintf(msg,"unknown command: %s",cmd);
  }
  switch (err) {
  case 0:
    if (*msg) message(g,msg,MSG_WINDOW);
    break;
  case E_ERROR:   
    message(g,msg,MSG_WARN);
    break;
  case E_NOTINIT:
    message(g,"Andor not initialized",MSG_WARN);
    break; 
  case E_RUNNING:
    message(g,"data acquisition running",MSG_WARN);
    break; 
  case E_MISSPAR: 
    message(g,"missing parameter",MSG_WARN);
    break;
  case E_NOTACQ:  
    message(g,"Andor not acquiring",MSG_WARN);
    break;
  case E_NOTIMP:  
    message(g,"command not implemented",MSG_WARN);
    break;
  default: 
    fprintf(stderr,"unknown error code %d\n",err);
    break;
  } 
}

/* ---------------------------------------------------------------- */

static void* run_setup(void* param)
{
  Guider *g = (Guider*)param;
  char   buf[128];
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p): %s\n",PREFUN,param,g->name);
#endif
  assert(!done);                       /* GUI up and running */
  assert(g->andor->handle >= 0);       /* connected */
  assert(g->andor->tid == 0);

  g->init_flag = -1;                   /* init running */
  message(g,PREFUN,MSG_INFO);

  strcpy(g->status.serial,g->andor->serialNumber);
  // IDEA if (!err) CBX_SetMainWindowName(&mwin,g->status.serial);

  long err = andor_setup(g->andor,baseD,baseB);
  if (!err) {
    assert(g->andor->aoiW == baseD/baseB);
    assert(g->andor->aoiH == baseD/baseB);
    g->status.binning = baseB;
    g->status.exptime = g->andor->expTime;
  }
  if (err) {                           /* IDEA just close GUI */
    g->init_flag = 0;                  /* there's no redemption :-( */
    sprintf(buf,"err= %ld",err);
    message(g,buf,MSG_ERROR);
    assert(g->andor->err == err);
  } else {
    g->init_flag = 1;
#ifdef SIM_ONLY
    sprintf(buf,"simfwhm %.3f",0.7/g->px);
    andor_sim(g->andor,buf,NULL);
    sprintf(buf,"simmode %d",g->shmode);
    andor_sim(g->andor,buf,NULL);
#endif
    set_exptime(g,0);                  /* update GUI */
    do_start(g,0);
    thread_detach(run_housekeeping,(void*)g);
  }

  return (void*)err;
}

/* ---------------------------------------------------------------- */

// TODO simplify run_init/wait

static void* run_wait(void* param)
{
  int  i;
  long err=0;
  char buf[128],dots[32]="."; /* IDEA limit dots to width of message box */
  Guider *g = (Guider*)param;
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p): %s\n",PREFUN,param,g->name);
#endif
  assert(!done);                       /* GUI up and running */
  assert(g->andor->tid == 0);

  g->init_flag = 0;
  message(g,PREFUN,MSG_INFO);
  sleep(1);
  andor_disconnect(g->andor);
  message(g,dots,MSG_WINDOW);

  for (i=0; i<60; i++) {               /* minnow reboot 60-100 seconds */
    sleep(1); if ((i % 6) == 0) strcat(dots,"."); 
    message(g,dots,MSG_OVER);
  }
  for (i=0; i<30; i++) {               /* 60 seconds */
    sleep(2);
    err = andor_connect(g->andor);     /* just TCP/IP connection */
    if (!err) break;
    strcat(dots,"."); message(g,dots,MSG_OVER);
  }

  if (!err) {
    message(g,"connecting to Andor ...",MSG_OVER);
    err = andor_open(g->andor);        /* actual USB connection */
  } else {
    message(g,"reboot timeout",MSG_ERROR);
  }
  if (!err) { 
    if (!strncmp(g->andor->serialNumber,"-E",2)) {
      sprintf(buf,"'%s' not connected to Andor",g->andor->host);
      message(g,buf,MSG_ERROR);
      err = E_andor_not_connected;
    }
  }
  if (!err) {
    err = (long)run_setup(g);
    if (err) {
      sprintf(buf,"err=%ld",err); message(g,buf,MSG_ERROR);
    }
  }
  g->init_flag = (err) ? 0 : 1;

  return (void*)err;
}

/* ---------------------------------------------------------------- */

static void* run_init(void* param)
{
  long err=0;
  char buf[512]; /* IDEA add dots[] */
  Guider *g = (Guider*)param;
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p): %s\n",PREFUN,param,g->name);
#endif
  assert(!done);                       /* GUI up and running */
 
  g->init_flag = -1;                   /* init running */
  message(g,PREFUN,MSG_INFO);

  if (!g->andor) {                     /* no host yet */
    run_scan(g);                       /* just scan this guider */
    g->init_flag = (g->andor) ? 1 : 0;
    return (void*)0;
  }

  if (g->loop_running) {               /* wait for loop to finish */
    do_stop(g,5000); msleep(1000);
  }
  if (g->andor->handle > 0) andor_close(g->andor);

  if (!err) {
    err = andor_open(g->andor);
    if (err && !strcmp(g->andor->host,"localhost")) { /* v0220 */
      sprintf(buf,"cannot connect to %s:%d",g->andor->host,g->andor->port);
      message(g,buf,MSG_WARN);
      int i = (g->andor->port - SERVER_PORT); /* v0221 */
      sprintf(buf,"nohup %s/andorserver -i%d &",getenv("HOME"),i);
      message(g,buf,MSG_INFO);
      system(buf);
      sleep(2);
      err = andor_open(g->andor);
    }
    if (err) {
      sprintf(buf,"cannot connect to %s:%d",g->andor->host,g->andor->port);
      message(g,buf,MSG_ERROR);
    } else {
      if (!strncmp(g->andor->serialNumber,"-E",2)) {
        sprintf(buf,"'%s' not connected to Andor",g->andor->host);
        message(g,buf,MSG_ERROR);
        err = E_andor_not_connected;
      }
    }
  }
  if (!err) {
    err = (long)run_setup(g);
    if (err) { char buf[128];
      sprintf(buf,"setup failed, err=%ld",err);
      message(g,buf,MSG_ERROR);
    }
  }
  if (!err) { 
    // TODO thread_detach(run_tcpip,(void*)g);
  }
  g->init_flag = (err) ? 0 : 1;

  return (void*)err;
}

/* ---------------------------------------------------------------- */

static void* run_tele(void* param)
{
  int err=0;
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p)\n",PREFUN,param);
#endif

  err = telio_open(2);
  if (!err) { int i;
    for (i=0; i<n_guiders; i++ ) { 
      Guider *g = guiders[i];
      if (g->pamode) err = set_pa(g,g->pa,1);
      if (err) break;
    }
    telio_close();
    // printf("elev=%.3f, para=%.3f\n",elev,para);
  }

  return (void*)(long)err;
}

/* ---------------------------------------------------------------- */

static void* run_scan_host(void* param)
{
  AndorIdentity *id = (AndorIdentity*)param;
  char buf[128];
  int  found=0;
#if (DEBUG > 0)
  fprintf(stderr,"%s(%s:%d)\n",PREFUN,id->host,id->port);
#endif

  AndorStruct *a = andor_create(id->host,id->port);
  int err = andor_open(a);
  if (!err) { int i;                   /* connected */
    for (i=0; i<n_guiders; i++) {
      if (guiders[i]->andor) continue;
      if (!strcmp(ident[i].serialNumber,a->serialNumber)) { /* found */
#ifdef SIM_ONLY
        sleep(3);
#endif
        found = 1;
        sprintf(buf,"found %s at %s:%d",a->serialNumber,id->host,id->port);
        message(guiders[i],buf,MSG_OVER);
        pthread_mutex_lock(&scanMutex);
        assert(!guiders[i]->andor);    /* Note: scanning prints no guider */
        guiders[i]->andor = a;         /* ID into logfile -- we don't */
        pthread_mutex_unlock(&scanMutex);  /* yet who's who */
        thread_detach(run_setup,(void*)guiders[i]);
        // TODO thread_detach(run_tcpip,(void*)guiders[i]);
        // IDEA check if server is running -- if not this is kinda lockfile
        break;
      }
    }
    if (!found) andor_disconnect(a);
  }
  if (!found) andor_free(a); 
  free((void*)param);

  return (void*)0;
}

/* --- */

static void* run_scan(void* param)
{
  int  i,j,p,p1=SERVER_PORT,p2,n=0,timeout=20;
  char buf[128],dots[32]="."; /* IDEA limit dots to width of message box */
#if (DEBUG > 0)
  fprintf(stderr,"%s(%p)\n",PREFUN,param);
#endif

  for (i=0; i<n_guiders; i++) {        /* find guiders without host */
    if (param) { if (guiders[i] != param) continue; } /* this guider only */
    if (guiders[i]->andor) continue;
    sprintf(buf,"scanning network (%s)",ident[i].serialNumber);
    message(guiders[i],buf,MSG_INFO);
    message(guiders[i],dots,MSG_WINDOW);
    n++;
  }
  if (n == 0) printf("nothing to scan\n");
  if (n == 0) return (void*)0;         /* nothing to scan */

  for (i=0; *knownHosts[i]; i++) {     /* scan all known hosts */
    p2 = (strstr(knownHosts[i],"minnow")) ? SERVER_PORT : SERVER_PORT+2;
    for (p=p1; p<=p2; p++) {           /* scan all known ports */
      int thisOne=1;                   /* don't scan connected hosts */
      pthread_mutex_lock(&scanMutex);
      for (j=0; j<n_guiders; j++) {
        if (!guiders[j]->andor) continue;
        if (strcmp(guiders[j]->andor->host,knownHosts[i])) continue;
        if (guiders[j]->andor->port != p) continue;
        thisOne = 0; break;            /* host/port already connected */
      }
      pthread_mutex_unlock(&scanMutex);
      if (!thisOne) continue; 
      AndorIdentity *id = (AndorIdentity*)malloc(sizeof(AndorIdentity));
      strcpy(id->host,knownHosts[i]);
      id->port = p;
      strcpy(id->serialNumber,"");
      thread_detach(run_scan_host,id);
    } 
  }

  do {                                 /* wait for results */
    sleep(1); timeout--; strcat(dots,".");
    pthread_mutex_lock(&scanMutex);
    for (i=0,n=0; i<n_guiders; i++) {
      if (guiders[i]->andor) n++;
      else                   message(guiders[i],dots,MSG_OVER);
    }
    pthread_mutex_unlock(&scanMutex);
    if (n == n_guiders) break;         /* all guiders found */
  } while (timeout > 0);

  for (i=0; i<n_guiders; i++) {        /* notify missing */
    if (!guiders[i]->andor) { 
      sprintf(buf,"%s not found",ident[i].serialNumber);
      message(guiders[i],buf,MSG_ERROR);
    }
  }

  return (void*)0;
}

/* ---------------------------------------------------------------- */

static void* run_housekeeping(void* param)
{
  Guider *g=(Guider*)param;
  int  slpt=2,err;
  char buf[128];

  g->stabilized = 0;
  do {
    sleep(1);
    if (g->init_flag != 1) break;      /* will be restarted at _setup */
    assert(g->andor);
    if (g->andor->handle < 0) break;   /* not connected */
    if (cor_time(0) % slpt) continue;
    err = andor_temperature(g->andor);
    if (!err) {
      g->status.temp_ccd = g->andor->tempSensor;
      sprintf(g->tmbox.text,"%.2f",g->andor->tempSensor); 
      CBX_UpdateEditWindow(&g->tmbox);
      sprintf(buf,"temp=%.2f (%s)",g->andor->tempSensor,
                                   g->andor->tempStatus); 
      message(g,buf,MSG_WINDOW);
      g->stabilized = (strcmp(g->andor->tempStatus,"Stabilised")) ? 0 : 1;
      slpt++;
    } /* endif(!err) */
  } while (!g->stabilized);
  sprintf(buf,"%s() done",PREFUN);
#if (DEBUG > 0)
  fprintf(stderr,"%s\n",buf);
#endif
  message(g,buf,MSG_FLUSH);

  return (void*)0;
}

/* ---------------------------------------------------------------- */

static int send_frame(Guider* g,AndorFrame* frame)
{
  int err=0;

  int sock = TCPIP_CreateClientSocket(g->send_host,g->send_port,&err);
  if (!err) { Header head;
    head.dimx = frame->w;
    head.dimy = frame->h;
    head.seqn = (int)frame->seqNumber;
    head.flag = g->shmode;
    assert(!is_bigendian());           /* network assumed big endian */
    swap4(&head,sizeof(head)/sizeof(int)); /* htonl() */
    ssize_t n1 = send(sock,&head,sizeof(head),0);
    size_t size = frame->w * frame->h * sizeof(u_short);
    u_short *data = (u_short*)malloc(size);
    swab(frame->data,data,size);
    ssize_t n2 = send(sock,data,size,0);
    (void)close(sock);
    free((void*)data);
    if (n1+n2 != size+sizeof(Header)) err = -1;
  }
  return err;
}

/* ---------------------------------------------------------------- */

static void* run_cycle(void* param)
{
  int    err=0,dt=0,last=0,q_flag=0;
  double t1,t2,t3,nodata=2+MAX_EXPTIME;
  char buf[128];
  u_int seqNumber=0;
  pthread_t disp_tid=0;                 
  Guider *g = (Guider*)param;
#if (DEBUG > 0)
  fprintf(stderr,"%s: %s(%p)\n",__FILE__,PREFUN,param);
#endif

  AndorStruct *andor = g->andor;
  assert(andor);
  assert(g->init_flag == 1);

  g->loop_running = True;
  g->stop_flag = False;

  err = andor_cycle_start(andor);

  if (!err) {
    t1 = walltime(0);
    thread_create(run_display,g,&disp_tid);
  }
  sprintf(g->dtbox.text,"%d",0); CBX_UpdateEditWindow(&g->dtbox);

  if (!err) { AndorFrame *frame=NULL;
    t1 = t2 = walltime(0);
    while (!g->stop_flag) {
      msleep(350);
      g->status.temp_ccd = g->andor->tempSensor;
      if ((seqNumber % 2) == 0) {
        sprintf(g->tmbox.text,"%.2f",g->andor->tempSensor);
        CBX_UpdateEditWindow(&g->tmbox);
      }
      t3 = walltime(0);
      frame = andor_frame4reading(andor,seqNumber);
      if (frame) {                     /* a new frame has arrived */
        seqNumber = frame->seqNumber;
        if (g->send_flag && (t3-t2 >= g->send_flag)) {
          int err = send_frame(g,frame);
          printf("%s: sending frame %u to %s:%d, err=%d\n",
                 get_ut_timestr(buf,cor_time(0)),seqNumber,
                 g->send_host,g->send_port,err);
          if (g->send_flag < 0) g->send_flag = 0;
          t2 = t3-0.175;
        }
        andor_frame_release(andor,frame);
        sprintf(g->sqbox.text,"%u",seqNumber);
        CBX_UpdateEditWindow(&g->sqbox);
        sprintf(g->fpbox.text,"%.1f",my_round(g->andor->fps,1));
        CBX_UpdateEditWindow(&g->fpbox);
        if (dt != 0) { dt = last = 0;
          sprintf(g->dtbox.text,"%d",dt); CBX_UpdateEditWindow(&g->dtbox);
        }
        t1 = t3-0.175;
        nodata = 2+MAX_EXPTIME;
      } else {                         /* no new frame */
        dt = (int)floor(t3-t1);
        if (dt != last) {
          sprintf(g->dtbox.text,"%d",dt); CBX_UpdateEditWindow(&g->dtbox);
          last = dt;
        }
        if (dt > nodata) {
          message(g,"no data",MSG_WARN);
          nodata *= 1.5;
        }
      }
      if (g->update_flag) {            /* update GUI from guider */
        pthread_mutex_lock(&g->mutex); /* lock and latch */
        double fps = g->fps;
        double flux = g->flux;
        double ppix = g->ppix;
        double back = g->back;
        double fwhm = g->fwhm;
        double dx = g->dx;
        double dy = g->dy;
        g->update_flag = False;
        pthread_mutex_unlock(&g->mutex);
        sprintf(g->fgbox.text,"%.1f",my_round(fps,1));
        CBX_UpdateEditWindow(&g->fgbox);
        if (fwhm > 0) {                /* we have a valid measurement */
          sprintf(g->tcbox.text,"tc %7.0f",flux);
          CBX_UpdateEditWindow(&g->tcbox);
          sprintf(g->mxbox.text,"mx %7.0f",ppix);
          CBX_UpdateEditWindow(&g->mxbox);
          sprintf(g->bkbox.text,"bk %7.0f",back);
          CBX_UpdateEditWindow(&g->bkbox);
          sprintf(g->fwbox.text,"fw %7.2f",fwhm);
          CBX_UpdateEditWindow(&g->fwbox);
          sprintf(g->dxbox.text,"dx %7.2f",dx);
          CBX_UpdateEditWindow(&g->dxbox);
          sprintf(g->dybox.text,"dy %7.2f",dy);
          CBX_UpdateEditWindow(&g->dybox);
          graph_redraw(g->g_tc);
          graph_redraw(g->g_fw);
          graph_redraw(g->g_az);
          graph_redraw(g->g_el);
        } /* endif(fwhm) */
        if (q_flag != g->q_flag) {     /* quality flag */
          q_flag = g->q_flag;
          switch (q_flag) {
            case 0: g->gdbox.fg = app->black; break;
            case 1: g->gdbox.fg = app->yellow; break;
            case 2: g->gdbox.fg = app->red; break;
          }
          CBX_UpdateEditWindow(&g->gdbox);
        }
      } /* endif(update) */
    } /* endwhile(!stop_flag) */
    andor_cycle_stop(andor);
  } /* endif(!err) */
  g->loop_running = False;
  sprintf(g->fpbox.text,"%d",0); CBX_UpdateEditWindow(&g->fpbox);
  sprintf(g->dtbox.text,"-"); CBX_UpdateEditWindow(&g->dtbox);

  if (disp_tid) pthread_join(disp_tid,NULL);

  if (err) {
    sprintf(buf,"%s: err=%d",PREFUN,err);
    message(g,buf,MSG_ERROR);
  }
  fprintf(stderr,"%s() done\n",PREFUN);

  return (void*)(long)err;
}
 
/* ---------------------------------------------------------------- */

static void* run_display(void* param) 
{
  Guider *g = (Guider*)param;
  u_int seqNumber=0;
  double fps=1,t1,t2,slpt=0.350,tmax;
  QlTool *qltool = g->qltool;
  AndorStruct *andor = g->andor;
  assert(andor);
  assert(g->init_flag == 1);

  int gx = g->status.dimx/g->status.binning;
  int gy = g->status.dimy/g->status.binning;
  qltool_reset(qltool,gx,gy,1);

  t1 = walltime(0);
  while (g->loop_running) {
    AndorFrame *frame = andor_frame4reading(andor,seqNumber);
    if (frame) {
      seqNumber = frame->seqNumber; 
      qltool_update(qltool,frame->data);
      andor_frame_release(andor,frame);
      qltool_redraw(qltool,True);
      t2 = walltime(0); 
      fps = 0.7*fps + 0.3/(t2-t1);
      sprintf(g->fdbox.text,"%.1f",my_round(fps,1));
      CBX_UpdateEditWindow(&g->fdbox);
      t1 = t2;
    } // endif(frame)
    slpt = (throttle) ? 0.6*slpt + 0.4*slpt*(fps/throttle) : 0.02; /* v0066 */
    tmax = (throttle) ? 1.0/throttle : 0.35;
    slpt = fmax(0.02,fmin(tmax,fmax(slpt,g->status.exptime/3.0)));
    // printf("slpt=%.3f\n",slpt);
    msleep((int)(1000.0*slpt)); 
  } // endwhile(!stop)
  sprintf(g->fdbox.text,"%d",0); CBX_UpdateEditWindow(&g->fdbox);
  fprintf(stderr,"%s() done\n",PREFUN);

  return (void*)0;
}

/* ---------------------------------------------------------------- */

static void* run_write(void* param) 
{
  int  weStartedTheLoop=0;
  char buf[128];
  Guider *g = (Guider*)param;
  double fps=0,t1,t2;
  AndorStruct *andor = g->andor;
  FITSpars *st = &g->status;

  if (g->init_flag != 1) return (void*)0;
  if (!g->loop_running) {
    weStartedTheLoop = do_start(g,2000);
  }

  t1 = walltime(0);
  while (g->loop_running) {
    AndorFrame *frame = andor_frame4reading(andor,st->seqNumber);
    if (frame) {
      st->seqNumber = frame->seqNumber;
      st->start_int = frame->timeStamp;
      st->stop_int  = frame->timeStamp + st->exptime;
      int e = fits_save(frame->data,st);
      if (e) { 
        sprintf(buf,"FITS error=%d",e);
        message(g,buf,MSG_WARN);
      } else {
        sprintf(buf,"%s.fits",st->filename);
        message(g,buf,MSG_FLUSH);
        st->run += 1;
        if (st->run > MAX_FILENUM) {
          st->run = 1;
          message(g,"file number overflow",MSG_WARN);
        }
        sprintf(buf,FMT_RUN,g->num);
        put_long(setup_rc,buf,st->run);
      } 
      andor_frame_release(andor,frame);
      t2 = walltime(0);
      fps = 0.5*fps + 0.5/(t2-t1);
      printf("%s: got frame-%u, fps=%.1f\n",PREFUN,st->seqNumber,fps);
      t1 = t2;
      g->write_flag -= 1;
    } // endif(frame)
    if (g->write_flag <= 0) break;
    msleep(50);
  } // endwhile(loop-doing)

  if (weStartedTheLoop) do_stop(g,0);
  printf("%s() done\n",PREFUN);

  return (void*)0;
}

/* ---------------------------------------------------------------- */

static void* run_guider(void* param) 
{
  double t1,t2;
  double cx,cy,gx,gy,fwhm,back,flux,peak;
  double fit[6];
  int    ix,iy,ppix=0,vrad=0,npix=0;
  u_int  seqNumber=0;
  Pixel  *pbuf=NULL;
  Guider *g = (Guider*)param;
  QlTool *qltool = g->qltool;
  AndorStruct *andor = g->andor;

  g->fps = 1.0/g->status.exptime;
  gx = my_round(qltool->curx[QLT_BOX],1);  /* 0.1 pixel resolution */
  gy = my_round(qltool->cury[QLT_BOX],1);
  ix = (int)my_round(gx,0);
  iy = (int)my_round(gy,0);

  int err = telio_open(2);             /* NEW v0224 */
  if (err) { char buf[128];
    sprintf(buf,"connection to TCSIS failed (err=%d)",err);
    message(g,buf,MSG_ERROR);
  }

  t1 = walltime(0);
  while (g->loop_running && qltool->guiding) {
    msleep(20);
    AndorFrame *frame = andor_frame4reading(andor,seqNumber);
    if (frame) { double dx=0,dy=0;
      seqNumber = frame->seqNumber;
      fwhm = get_fwhm(frame->data,frame->w,frame->h,ix,iy,qltool->vrad,
                      qltool->enoise,qltool->egain,&back,&cx,&cy,&peak,&flux);
#if (DEBUG > 1)
      printf("back=%.1f, cx=%.1f, cy=%.1f, peak=%.1f, flux=%.0f fwhm=%.3f\n",
             back,cx,cy,peak,flux,fwhm);
#endif
      if (vrad != qltool->vrad) {      /* box size changed */
        vrad = qltool->vrad;
        npix =(1+2*vrad) * (1+2*vrad);
        pbuf = (Pixel*)realloc(pbuf,npix*sizeof(Pixel));
      }
      double ggx = my_round(qltool->curx[QLT_BOX],1);
      double ggy = my_round(qltool->cury[QLT_BOX],1);
      if ((gx != ggx) || (gy != ggy)) {
        gx = ggx; gy = ggy;
        printf("telescope dragging (%f,%f) !!!!\n",gx,gy); // always ?Shec
        ix = (int)my_round(gx,0); iy = (int)my_round(gy,0);
      }
      assert(npix); assert(pbuf);
      if (fwhm > 0) { int x,y,v,xx,yy,i=0; /* we have a valid estimate */
        ppix=0;                            /* peak pixel */
        for (xx=-vrad; xx<=vrad; xx++) {   /* loop over box */
          x = ix + xx;
          if (x < 0) continue;             /* outside frame */
          if (x >= frame->w) break;
          for (yy=-vrad; yy<=vrad; yy++) {
            y = iy + yy;
            if (y < 0) continue;
            if (y >= frame->h) break;
            pbuf[i].x = x;
            pbuf[i].y = y;
            v = frame->data[x+y*frame->w]; if (v > ppix) ppix = v;
            pbuf[i].z = (double)v;
            i++;
          }
        }
        assert(i <= npix);
        fit[0] = back;
        fit[1] = cx;
        fit[2] = cy;
        fit[3] = flux;
        fit[4] = fwhm/2.35482;
        fit[5] = peak;
        int itmax = 10+(int)(500.0*g->status.exptime);
        int it = ccbphot(pbuf,i,fit,itmax);
        // allow '2' -> '1' ?Shec
        g->q_flag = (it > itmax) ? ((g->q_flag==2) ? 2 : 1) : 0;
        back = fit[0];
        dx   = fit[1];
        dy   = fit[2];
        flux = fit[3];
        fwhm = fit[4] * g->px;
        peak = fit[5];
      } else {
        fprintf(stderr,"no star !!!!!!!!!!!!!!!!!!!!\n");
        g->q_flag = 2;
      } /* endif(fwhm) */
      andor_frame_release(andor,frame);
      t2 = walltime(0);
      pthread_mutex_lock(&g->mutex);
      g->fps = 0.8*g->fps + 0.2/(t2-t1);
      t1 = t2;
      if (g->q_flag < 2) { double azerr,elerr; /* we have valid measurement */
        if (qltool->guiding == -5) {   /* adjust guider box */
          gx = dx; ix = qltool->curx[QLT_BOX] = (int)my_round(gx,0);
          gy = dy; iy = qltool->cury[QLT_BOX] = (int)my_round(gy,0);
        }
        g->flux = flux;
        g->ppix = (double)ppix;
        g->back = back;  
        g->fwhm = fwhm;  
        g->dx = legacy*(dx-gx);        /* legacy pixels ?Shec */
        g->dy = legacy*(dy-gy);
        if (qltool->guiding < 0) graph_scale(g->g_tc,0,1.333*flux,0);
        graph_add1(g->g_tc,flux,0);
        graph_add1(g->g_fw,fwhm,0);
        rotate(g->px*(dx-gx),g->px*(dy-gy),g->pa,&azerr,&elerr);
        graph_add1(g->g_az,azerr,0);
        graph_add1(g->g_el,elerr,0);
        const double fdge=0.35;  // TODO reduce sensitiviy if q_flag==1?
        if (qltool->guiding < 0) {     /* first pass -- no derivative */
          g->azg = fdge * g->sens * azerr;
          g->elg = fdge * g->sens * elerr;
        } else {                       /* use derivative */
          g->azg  = azerr + ((andor->rolling + 1.0) * (azerr - g->azerp));
          g->azg *= fdge * g->sens;
          g->elg  = elerr + ((andor->rolling + 1.0) * (elerr - g->elerp));
          g->elg *= fdge * g->sens;
        }
#ifdef TEST_GUIDER
        graph_add(g->g_az,1,g->azg,0);
        graph_add(g->g_el,1,g->elg,0);
#endif
        g->azerp = azerr;              /* store last error */
        g->elerp = elerr;
        qltool->guiding = abs(qltool->guiding);
        if ((qltool->guiding == 3) || (qltool->guiding == 5)) {
          telio_aeg(g->azg,g->elg);    /* NEW v0224 */
    printf("%6.3f (%6.3f) %6.3f (%6.3f)\n",g->azg,azerr,g->elg,elerr); //xxx
        }
      } else {
        g->azg = g->elg = 0;
      } // endif(q_flag)
      g->update_flag = True;
      pthread_mutex_unlock(&g->mutex);
    } // endif(frame)
  } // endwhile(loop-doing)
  qltool->guiding = 0;

  free((void*)pbuf);
  telio_close();

  sprintf(g->fgbox.text,"%d",0); CBX_UpdateEditWindow(&g->fgbox);
  printf("%s(%s) -- stopped\n",PREFUN,g->name);

  return (void*)0;
}

/* ---------------------------------------------------------------- */

static int set_pa(Guider* g,double f,int fromTCS) /* position angle */ 
{
  int err=0;
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p,%.1f,%d)\n",PREFUN,g,f,fromTCS);
#endif

  if (fromTCS) { float el,pa;
    err = telio_geo(NULL,NULL,NULL,NULL,&el,&pa); /* am,ra,re,az,el,pa */
    if (!err) {
      f = -1.0*(120.0+el);             /* TODO parity from .ini */
      elev = el;                       /* telescope elevation */
      para = pa;                       /* parallactic angle */
    }
  }
  if (f == 0.0) f = 360.0;
  while (f >  360.0) f -= 360.0;
  while (f < -360.0) f += 360.0;
  g->pa = f;
  sprintf(g->pabox.text,"pa %4.0f",my_round(g->pa,0));
#ifndef PA_BOX
  g->pabox.fg = (g->pamode) ? app->green : app->black;
#endif
  CBX_UpdateEditWindow(&g->pabox);
  redraw_compass(g);

  return err;
}

/* ---------------------------------------------------------------- */

static void set_sens(Guider* g,double f)
{
#if (DEBUG > 1)
  fprintf(stderr,"%s(%s,%f)\n",PREFUN,g->name,f);
#endif

  g->sens = fmin(4.0,fmax(0.1,f));
  sprintf(g->snbox.text,"sn %4.1f",g->sens);
  CBX_UpdateEditWindow(&g->snbox);
}

/* ---------------------------------------------------------------- */

static void set_mm(Guider* g,int m)
{
  switch (g->num) { 
  case MODE_PR: case MODE_SH:            /* PR,SH guiders */
    switch (m) {
    case 1: case  2: case 3:             /* allowed mouse modes */
      g->msmode = m;
      break;
    }
    break;
  case MODE_SV:                          /* SV guider */
    switch (m) {
    case 1: case -2: case 3:             /* allowed mouse modes */
      g->msmode = m;
      break;
    }
    break;
  }
  if (g->msmode != m) {
    message(g,"invalid mouse mode",MSG_WARN);
  }
  sprintf(g->mmbox.text,"mm %4d",g->msmode);
  CBX_UpdateEditWindow(&g->mmbox);
}

/* ---------------------------------------------------------------- */

static int set_exptime(Guider* g,float t)
{
  int err=0; 

  if (t > 0) {
    double tmin = (g->andor->aoiW*g->andor->aoiH)/90.0e6;
    tmin = 0.01*ceil(tmin/0.01);
    g->status.exptime = (float)fmin(MAX_EXPTIME,fmax(tmin,t));
  }
  sprintf(g->tfbox.text,"%.*f",(g->status.exptime < 10) ? 2 : 1,
          g->status.exptime);
  CBX_ResetEditWindow(&g->tfbox);

  if (t > 0) {
    err = andor_exptime(g->andor,g->status.exptime);
    if (err) { char buf[128];
      g->status.exptime = (float)g->andor->expTime;
      sprintf(buf,"sending 'ExpTime' failed, err=%d",err);
      message(g,buf,MSG_ERROR);
      set_exptime(g,0);                /* just update GUI */
    }
  }
  return err;
}

/* ---------------------------------------------------------------- */

void message(const void *param,const char* text,int flags)
{
  Guider *g=NULL;
  int  i=0;
  char tstr[32];
  static FILE* fp=NULL;
#if (DEBUG > 1)
  fprintf(stderr,"%s(%p,%s,%x)\n",PREFUN,param,text,flags);
#endif
  assert(text);

  if (param) {
    for (i=0; i<n_guiders; i++) {      /* source */
      if (param == guiders[i]) { g = guiders[i]; break; }
      if (param == guiders[i]->andor) { g = guiders[i]; break; }
    } 
  }

  if (flags & MSG_FILE) { char who[64];  /* write to file */
    pthread_mutex_lock(&mesgMutex); 
    if (!fp) fp = fopen(logfile,"a");
    if (fp) {
      strcpy(who,(g) ? g->name : P_TITLE);
      if      (flags & MSG_RED)     strcat(who,"-ERROR");
      else if (flags & MSG_YELLOW)  strcat(who,"-WARNING");
      get_ut_datestr(tstr,cor_time(0));
      fprintf(fp,"%s - %s: %s\n",tstr,who,text);
      if (flags & MSG_FLUSH) fflush(fp);
    } 
    pthread_mutex_unlock(&mesgMutex); 
  }

  if (!g) return;
  if (!(flags & MSG_WINDOW)) return;
  assert(mwin.disp);

  get_ut_timestr(tstr,cor_time(0));
  if (flags & MSG_OVERWRITE) {         /* overwrite */
    sprintf(g->msbox[0].text,"%s: %s",tstr,text);
    CBX_UpdateEditWindow(&g->msbox[0]);
    return;
  }
 
  pthread_mutex_lock(&mesgMutex); 
  for (i=n_msg-1; i>=0; i--) {         /* prepend line */
    if (i == 0) {
      sprintf(g->msbox[i].text,"%s: %s",tstr,text);
      if      (flags & MSG_RED)    CBX_ChangeLed(&g->led[i],app->red); 
      else if (flags & MSG_YELLOW) CBX_ChangeLed(&g->led[i],app->yellow); 
      else                         CBX_ChangeLed(&g->led[i],app->green); 
    } else {
      strcpy(g->msbox[i].text,g->msbox[i-1].text);
      CBX_ChangeLed(&g->led[i],g->led[i-1].bg);
    }
    CBX_UpdateEditWindow(&g->msbox[i]);
  }
  pthread_mutex_unlock(&mesgMutex); 
}

/* ---------------------------------------------------------------- */

static int check_datapath(char* path,int interactive)
{
  char buffer[1024],cmd[1024],*h;
#if (DEBUG > 1)
  fprintf(stderr,"%s(%s,%d)\n",PREFUN,path,interactive);
#endif

  cut_spaces(path);
  if (*path == '\0') return 0;

  if (path[0] != '/') {
    if (interactive) {
      sprintf(buffer,"invalid path (%s), must be absolute ('/...')",path);
      CBX_MessageBox(P_TITLE,buffer);
    }
    *path = '\0';
    return 0;
  }

  if (chdir(path) == -1) {
    if (interactive) {
      if (CBX_AlertBox(path,"does not exist - create new directory ?")) {
        strcpy(buffer,path); path[0]='\0';
        h=strtok(buffer,"/");
        do {
          strcat(path,"/"); strcat(path,h);
          if (chdir(path) == -1) {       /* path does not exist */
            sprintf(cmd,"mkdir %s",path);
            (void)system(cmd);
          }
        } while ((h=strtok(NULL,"/")) != NULL);
      } else {
        strcpy(path,genv2("HOME","tmp"));
        CBX_MessageBox(P_TITLE ": set DataPath to",path);
      }
      CBX_ClearAutoQueue(&mwin);
    } else {
      *path = '\0';
    }
  }
  if (*path) next_disk = 0;

  return (*path) ? 1 : 0;
}

/* ---------------------------------------------------------------- */

static void my_shutdown(Bool force)
{
  int i;

  if (!force) {
    force = True;
    for (i=0; i<n_guiders; i++) {
      if (guiders[i]->loop_running) force = False;
    }
  }
  if (force) { Bool wait=False;
    for (i=0; i<n_guiders; i++) {
      if (guiders[i]->loop_running) {
        wait = True;
        do_stop(guiders[i],0);
      }
    }
    if (wait) msleep(1000);
    done = True;
  }
}

/* ---------------------------------------------------------------- */

static int handle_tcpip(const char* command,char* answer,size_t buflen)
{
  int  r=0;
  char cmd[128]="",par[128]="";

  strcpy(answer,"-Einvalid command\n");
  if (strlen(command) >= sizeof(cmd)) return 0;

  int n = sscanf(command,"%s %s",cmd,par);
  if (n < 1) return 0;

  if (!strcasecmp(cmd,"version")) {
    sprintf(answer,"%s\n",P_VERSION);
  } else 
  if (!strcasecmp(cmd,"quit")) {
    r = 1;
    strcpy(answer,"OK\n");
  } else {
    strcpy(answer,"-Eunknown command\n");
  }

  return r;
}

/* --- */

static int receive_string(int sock,char* string,size_t buflen)
{
  int     i=0;
  ssize_t r;

  do {
    r = recv(sock,&string[i],1,0);
    if (r != 1) break;
    if (string[i] == '\n') break;      /* <LF> */
    if (string[i] == '\r') break;      /* <CR> */
    i++;
  } while (i < buflen-1);
  string[i] = '\0';                    /* overwrite term-char */

  return r;
}

/* --- */

static void* run_tcpip(void* param)
{
  Guider *g = (Guider*)param;
  int    port = TCPIP_PORT + g->num;
  int    err,msgsock,rval,r=0;
  char   cmd[128],buf[128];
  struct  sockaddr sadr;
  socklen_t size=sizeof(sadr);
#if (DEBUG > 0)
  fprintf(stderr,"%s: %s(%d)\n",__FILE__,PREFUN,port);
#endif

  if (param) return (void*)0;         // TODO 

  int sock = TCPIP_CreateServerSocket(port,&err);
  if (err) { 
    sprintf(buf,"TCPIP-Server failed, err=%d",err);
    message(g,buf,MSG_WARN);
    return (void*)(long)err;
  }
  sprintf(buf,"TCPIP-Server listening (%s,%d)",g->name,port);
  message(g,buf,MSG_FLUSH);

  while (!done) {
    msgsock = accept(sock,&sadr,&size);
    sprintf(buf,"connection accepted from %u.%u.%u.%u",
        (u_char)sadr.sa_data[2],(u_char)sadr.sa_data[3],
        (u_char)sadr.sa_data[4],(u_char)sadr.sa_data[5]);
    message(g,buf,MSG_FLUSH);
#ifdef SO_NOSIGPIPE                    /* 2017-07-05 */
    int on=1; (void)setsockopt(msgsock,SOL_SOCKET,SO_NOSIGPIPE,&on,sizeof(on));
#endif
    do {
      rval = receive_string(msgsock,cmd,sizeof(cmd));
      if (rval > 0) {
        fprintf(stdout,"%s(): received '%s'\n",PREFUN,cmd);
        r = handle_tcpip(cmd,buf,sizeof(buf));
        send(msgsock,buf,strlen(buf),0);
      } else {
        fprintf(stdout,"%s(): rval=%d\n",PREFUN,rval);
      }
      if (r) break;
    } while (rval > 0);                /* loop while there's something */
    (void)close(msgsock);
  } /* while(!done) */

  (void)close(sock);
#if (DEBUG > 0)
  fprintf(stderr,"%s(%d) done\n",PREFUN,port);
#endif

  return (void*)0;
}

/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */
/* ---------------------------------------------------------------- */

